# Copyright 2015 Mirantis Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
from unittest import mock

import ddt
from oslo_config import cfg

from manila.common import constants as const
from manila import exception
import manila.share.configuration
from manila.share.drivers import helpers
from manila import test
from manila.tests import fake_compute
from manila.tests import fake_utils
from manila.tests.share.drivers import test_generic


CONF = cfg.CONF


@ddt.ddt
class NFSHelperTestCase(test.TestCase):
    """Test case for NFS helper."""

    def setUp(self):
        super(NFSHelperTestCase, self).setUp()
        fake_utils.stub_out_utils_execute(self)
        self.fake_conf = manila.share.configuration.Configuration(None)
        self._ssh_exec = mock.Mock(return_value=('', ''))
        self._execute = mock.Mock(return_value=('', ''))
        self._helper = helpers.NFSHelper(self._execute, self._ssh_exec,
                                         self.fake_conf)
        ip = '10.254.0.3'
        self.server = fake_compute.FakeServer(
            ip=ip, public_address=ip, instance_id='fake_instance_id')
        self.share_name = 'fake_share_name'

    def test_init_helper(self):

        # mocks
        self.mock_object(
            self._helper, '_ssh_exec',
            mock.Mock(side_effect=exception.ProcessExecutionError(
                stderr='command not found')))

        # run
        self.assertRaises(exception.ManilaException,
                          self._helper.init_helper, self.server)

        # asserts
        self._helper._ssh_exec.assert_called_once_with(
            self.server, ['sudo', 'exportfs'])

    def test_init_helper_log(self):

        # mocks
        self.mock_object(
            self._helper, '_ssh_exec',
            mock.Mock(side_effect=exception.ProcessExecutionError(
                stderr='fake')))

        # run
        self._helper.init_helper(self.server)

        # asserts
        self._helper._ssh_exec.assert_called_once_with(
            self.server, ['sudo', 'exportfs'])

    @ddt.data(
        {"server": {"public_address": "1.2.3.4"}, "version": 4},
        {"server": {"public_address": "1001::1002"}, "version": 6},
        {"server": {"public_address": "1.2.3.4", "admin_ip": "5.6.7.8"},
         "version": 4},
        {"server": {"public_address": "1.2.3.4", "ip": "9.10.11.12"},
         "version": 4},
        {"server": {"public_address": "1001::1001", "ip": "1001::1002"},
         "version": 6},
        {"server": {"public_address": "1001::1002", "admin_ip": "1001::1002"},
         "version": 6},
        {"server": {"public_addresses": ["1001::1002"]}, "version": 6},
        {"server": {"public_addresses": ["1.2.3.4", "1001::1002"]},
         "version": {"1.2.3.4": 4, "1001::1002": 6}},
    )
    @ddt.unpack
    def test_create_exports(self, server, version):
        result = self._helper.create_exports(server, self.share_name)

        expected_export_locations = []
        path = os.path.join(CONF.share_mount_path, self.share_name)
        service_address = server.get("admin_ip", server.get("ip"))
        version_copy = version

        def convert_address(address, version):
            if version == 4:
                return address
            return "[%s]" % address

        if 'public_addresses' in server:
            pairs = list(map(lambda addr: (addr, False),
                             server['public_addresses']))
        else:
            pairs = [(server['public_address'], False)]

        service_address = server.get("admin_ip", server.get("ip"))
        if service_address:
            pairs.append((service_address, True))

        for ip, is_admin in pairs:
            if isinstance(version_copy, dict):
                version = version_copy.get(ip)

            expected_export_locations.append({
                "path": "%s:%s" % (convert_address(ip, version), path),
                "is_admin_only": is_admin,
                "metadata": {
                    "export_location_metadata_example": "example",
                },
            })
        self.assertEqual(expected_export_locations, result)

    @ddt.data(const.ACCESS_LEVEL_RW, const.ACCESS_LEVEL_RO)
    def test_update_access(self, access_level):
        expected_mount_options = '%s,no_subtree_check,no_root_squash'
        self.mock_object(self._helper, '_sync_nfs_temp_and_perm_files')
        local_path = os.path.join(CONF.share_mount_path, self.share_name)
        exec_result = ' '.join([local_path, '2.2.2.3'])
        self.mock_object(self._helper, '_ssh_exec',
                         mock.Mock(return_value=(exec_result, '')))
        access_rules = [
            test_generic.get_fake_access_rule('1.1.1.1', access_level),
            test_generic.get_fake_access_rule('2.2.2.2', access_level),
            test_generic.get_fake_access_rule('2.2.2.3', access_level)]
        add_rules = [
            test_generic.get_fake_access_rule('2.2.2.2', access_level),
            test_generic.get_fake_access_rule('2.2.2.3', access_level),
            test_generic.get_fake_access_rule('5.5.5.0/24', access_level)]
        delete_rules = [
            test_generic.get_fake_access_rule('3.3.3.3', access_level),
            test_generic.get_fake_access_rule('4.4.4.4', access_level, 'user'),
            test_generic.get_fake_access_rule('0.0.0.0/0', access_level)]
        self._helper.update_access(self.server, self.share_name, access_rules,
                                   add_rules=add_rules,
                                   delete_rules=delete_rules)
        local_path = os.path.join(CONF.share_mount_path, self.share_name)
        self._helper._ssh_exec.assert_has_calls([
            mock.call(self.server, ['sudo', 'exportfs']),
            mock.call(self.server, ['sudo', 'exportfs', '-u',
                                    ':'.join(['3.3.3.3', local_path])]),
            mock.call(self.server, ['sudo', 'exportfs', '-u',
                                    ':'.join(['*',
                                              local_path])]),
            mock.call(self.server, ['sudo', 'exportfs', '-o',
                                    expected_mount_options % access_level,
                                    ':'.join(['2.2.2.2', local_path])]),
            mock.call(self.server, ['sudo', 'exportfs', '-o',
                                    expected_mount_options % access_level,
                                    ':'.join(['5.5.5.0/24',
                                              local_path])]),
        ])
        self._helper._sync_nfs_temp_and_perm_files.assert_has_calls([
            mock.call(self.server), mock.call(self.server)])

    @ddt.data({'access': '10.0.0.1', 'result': '10.0.0.1'},
              {'access': '10.0.0.1/32', 'result': '10.0.0.1'},
              {'access': '10.0.0.0/24', 'result': '10.0.0.0/24'},
              {'access': '1001::1001', 'result': '[1001::1001]'},
              {'access': '1001::1000/128', 'result': '[1001::1000]'},
              {'access': '1001::1000/124', 'result': '[1001::1000]/124'})
    @ddt.unpack
    def test__get_parsed_address_or_cidr(self, access, result):
        self.assertEqual(result,
                         self._helper._get_parsed_address_or_cidr(access))

    @ddt.data('10.0.0.265', '10.0.0.1/33', '1001::10069', '1001::1000/129')
    def test__get_parsed_address_or_cidr_with_invalid_access(self, access):
        self.assertRaises(ValueError,
                          self._helper._get_parsed_address_or_cidr,
                          access)

    def test_update_access_invalid_type(self):
        access_rules = [test_generic.get_fake_access_rule(
            '2.2.2.2', const.ACCESS_LEVEL_RW, access_type='fake'), ]
        self.assertRaises(
            exception.InvalidShareAccess,
            self._helper.update_access,
            self.server,
            self.share_name,
            access_rules,
            [],
            [])

    def test_update_access_invalid_level(self):
        access_rules = [test_generic.get_fake_access_rule(
            '2.2.2.2', 'fake_level', access_type='ip'), ]
        self.assertRaises(
            exception.InvalidShareAccessLevel,
            self._helper.update_access,
            self.server,
            self.share_name,
            access_rules,
            [],
            [])

    @ddt.data({'access_to': 'lala', 'access_type': 'user'},
              {'access_to': '203.0.113.29'},
              {'access_to': '2001:0DB8:7d18:c63e:5f0a:871f:83b8:d244',
               'access_level': 'ro'})
    @ddt.unpack
    def test_update_access_delete_invalid_rule(
            self, access_to, access_level='rw', access_type='ip'):
        mount_path = '%s:/shares/%s' % (access_to, self.share_name)
        if access_type == 'ip':
            self._helper._get_parsed_address_or_cidr = mock.Mock(
                return_value=access_to)
            not_found_msg = (
                "exportfs: Could not find '%s' to unexport.\n" % mount_path
            )
            exc = exception.ProcessExecutionError
            self.mock_object(
                self._helper,
                '_ssh_exec',
                mock.Mock(side_effect=[(0, 0), exc(stderr=not_found_msg)]))

        delete_rules = [
            test_generic.get_fake_access_rule(access_to,
                                              access_level,
                                              access_type),
        ]
        self.mock_object(self._helper, '_sync_nfs_temp_and_perm_files')

        self._helper.update_access(self.server, self.share_name, [],
                                   [], delete_rules)

        if access_type == 'ip':
            self._helper._ssh_exec.assert_has_calls([
                mock.call(self.server, ['sudo', 'exportfs']),
                mock.call(self.server,
                          ['sudo', 'exportfs', '-u', mount_path])])
        self._helper._sync_nfs_temp_and_perm_files.assert_called_with(
            self.server)

    def test_get_host_list(self):
        fake_exportfs = ('/shares/share-1\n\t\t20.0.0.3\n'
                         '/shares/share-1\n\t\t20.0.0.6\n'
                         '/shares/share-2\n\t\t10.0.0.2\n'
                         '/shares/share-2\n\t\t10.0.0.5\n'
                         '/shares/share-3\n\t\t30.0.0.4\n'
                         '/shares/share-3\n\t\t30.0.0.7\n')
        expected = ['20.0.0.3', '20.0.0.6']
        result = self._helper.get_host_list(fake_exportfs, '/shares/share-1')
        self.assertEqual(expected, result)

    @ddt.data({"level": const.ACCESS_LEVEL_RW, "ip": "1.1.1.1",
               "expected": "1.1.1.1"},
              {"level": const.ACCESS_LEVEL_RO, "ip": "1.1.1.1",
               "expected": "1.1.1.1"},
              {"level": const.ACCESS_LEVEL_RW, "ip": "fd12:abcd::10",
               "expected": "[fd12:abcd::10]"},
              {"level": const.ACCESS_LEVEL_RO, "ip": "fd12:abcd::10",
               "expected": "[fd12:abcd::10]"})
    @ddt.unpack
    def test_update_access_recovery_mode(self, level, ip, expected):
        expected_mount_options = '%s,no_subtree_check,no_root_squash'
        access_rules = [test_generic.get_fake_access_rule(
            ip, level), ]
        self.mock_object(self._helper, '_sync_nfs_temp_and_perm_files')
        self.mock_object(self._helper, 'get_host_list',
                         mock.Mock(return_value=[ip]))
        self._helper.update_access(self.server, self.share_name, access_rules,
                                   [], [])
        local_path = os.path.join(CONF.share_mount_path, self.share_name)
        self._ssh_exec.assert_has_calls([
            mock.call(self.server, ['sudo', 'exportfs']),
            mock.call(
                self.server, ['sudo', 'exportfs', '-u',
                              ':'.join([expected,
                                        local_path])]),
            mock.call(self.server, ['sudo', 'exportfs', '-o',
                                    expected_mount_options % level,
                                    ':'.join([expected, local_path])]),
        ])
        self._helper._sync_nfs_temp_and_perm_files.assert_called_with(
            self.server)

    def test_sync_nfs_temp_and_perm_files(self):
        self._helper._sync_nfs_temp_and_perm_files(self.server)
        self._helper._ssh_exec.assert_has_calls(
            [mock.call(self.server, mock.ANY) for i in range(1)])

    @ddt.data('/foo/bar', '5.6.7.8:/bar/quuz', '5.6.7.9:/foo/quuz',
              '[1001::1001]:/foo/bar', '[1001::1000]/:124:/foo/bar')
    def test_get_exports_for_share_single_ip(self, export_location):
        server = dict(public_address='1.2.3.4')

        result = self._helper.get_exports_for_share(server, export_location)

        path = export_location.split(':')[-1]
        expected_export_locations = [
            {"is_admin_only": False,
             "path": "%s:%s" % (server["public_address"], path),
             "metadata": {"export_location_metadata_example": "example"}}
        ]
        self.assertEqual(expected_export_locations, result)

    @ddt.data('/foo/bar', '5.6.7.8:/bar/quuz', '5.6.7.9:/foo/quuz')
    def test_get_exports_for_share_multi_ip(self, export_location):
        server = dict(public_addresses=['1.2.3.4', '1.2.3.5'])

        result = self._helper.get_exports_for_share(server, export_location)

        path = export_location.split(':')[-1]
        expected_export_locations = list(map(
            lambda addr: {
                "is_admin_only": False,
                "path": "%s:%s" % (addr, path),
                "metadata": {"export_location_metadata_example": "example"}
            },
            server['public_addresses'])
        )
        self.assertEqual(expected_export_locations, result)

    @ddt.data(
        {'public_address_with_suffix': 'foo'},
        {'with_prefix_public_address': 'bar'},
        {'with_prefix_public_address_and_with_suffix': 'quuz'}, {})
    def test_get_exports_for_share_with_error(self, server):
        export_location = '1.2.3.4:/foo/bar'

        self.assertRaises(
            exception.ManilaException,
            self._helper.get_exports_for_share, server, export_location)

    @ddt.data('/foo/bar', '5.6.7.8:/foo/bar', '5.6.7.88:fake:/foo/bar',
              '[1001::1002]:/foo/bar', '[1001::1000]/124:/foo/bar')
    def test_get_share_path_by_export_location(self, export_location):
        result = self._helper.get_share_path_by_export_location(
            dict(), export_location)

        self.assertEqual('/foo/bar', result)

    @ddt.data(
        ('/shares/fake_share1\n\t\t1.1.1.10\n'
         '/shares/fake_share2\n\t\t1.1.1.16\n'
         '/shares/fake_share3\n\t\t<world>\n'
         '/mnt/fake_share1 1.1.1.11', False),
        ('/shares/fake_share_name\n\t\t1.1.1.10\n'
         '/shares/fake_share_name\n\t\t1.1.1.16\n'
         '/shares/fake_share_name\n\t\t<world>\n'
         '/mnt/fake_share1\n\t\t1.1.1.11', True),
        ('/mnt/fake_share_name\n\t\t1.1.1.11\n'
         '/shares/fake_share_name\n\t\t1.1.1.10\n'
         '/shares/fake_share_name\n\t\t1.1.1.16\n'
         '/shares/fake_share_name\n\t\t<world>\n', True))
    @ddt.unpack
    def test_disable_access_for_maintenance(self, output, hosts_match):
        fake_maintenance_path = "fake.path"
        self._helper.configuration.share_mount_path = '/shares'
        local_path = os.path.join(self._helper.configuration.share_mount_path,
                                  self.share_name)

        def fake_ssh_exec(*args, **kwargs):
            if 'exportfs' in args[1] and '-u' not in args[1]:
                return output, ''
            else:
                return '', ''

        self.mock_object(self._helper, '_ssh_exec',
                         mock.Mock(side_effect=fake_ssh_exec))

        self.mock_object(self._helper, '_sync_nfs_temp_and_perm_files')
        self.mock_object(self._helper, '_get_maintenance_file_path',
                         mock.Mock(return_value=fake_maintenance_path))

        self._helper.disable_access_for_maintenance(
            self.server, self.share_name)

        self._helper._ssh_exec.assert_any_call(
            self.server,
            ['cat', const.NFS_EXPORTS_FILE,
             '|', 'grep', self.share_name,
             '|', 'sudo', 'tee', fake_maintenance_path]
        )
        self._helper._ssh_exec.assert_has_calls([
            mock.call(self.server, ['sudo', 'exportfs']),
        ])

        if hosts_match:
            self._helper._ssh_exec.assert_has_calls([
                mock.call(self.server,
                          ['sudo', 'exportfs', '-u',
                           '"{}"'.format(':'.join(['1.1.1.10', local_path]))]),
                mock.call(self.server,
                          ['sudo', 'exportfs', '-u',
                           '"{}"'.format(':'.join(['1.1.1.16', local_path]))]),
                mock.call(self.server,
                          ['sudo', 'exportfs', '-u',
                           '"{}"'.format(':'.join(['<world>', local_path]))]),
            ])

        self._helper._sync_nfs_temp_and_perm_files.assert_called_once_with(
            self.server
        )

    def test_restore_access_after_maintenance(self):
        fake_maintenance_path = "fake.path"
        self.mock_object(self._helper, '_get_maintenance_file_path',
                         mock.Mock(return_value=fake_maintenance_path))
        self.mock_object(self._helper, '_ssh_exec')

        self._helper.restore_access_after_maintenance(
            self.server, self.share_name)

        self._helper._ssh_exec.assert_called_once_with(
            self.server,
            ['cat', fake_maintenance_path,
             '|', 'sudo', 'tee', '-a', const.NFS_EXPORTS_FILE,
             '&&', 'sudo', 'exportfs', '-r', '&&', 'sudo', 'rm', '-f',
             fake_maintenance_path]
        )


@ddt.ddt
class CIFSHelperIPAccessTestCase(test.TestCase):
    """Test case for CIFS helper with IP access."""

    def setUp(self):
        super(CIFSHelperIPAccessTestCase, self).setUp()
        self.server_details = {'instance_id': 'fake',
                               'public_address': '1.2.3.4', }
        self.share_name = 'fake_share_name'
        self.fake_conf = manila.share.configuration.Configuration(None)
        self._ssh_exec = mock.Mock(return_value=('', ''))
        self._execute = mock.Mock(return_value=('', ''))
        self._helper = helpers.CIFSHelperIPAccess(self._execute,
                                                  self._ssh_exec,
                                                  self.fake_conf)
        self.access = dict(
            access_level=const.ACCESS_LEVEL_RW,
            access_type='ip',
            access_to='1.1.1.1')

    def test_init_helper(self):
        self._helper.init_helper(self.server_details)
        self._helper._ssh_exec.assert_called_once_with(
            self.server_details,
            ['sudo', 'net', 'conf', 'list'],
        )

    def test_create_export_share_does_not_exist(self):
        def fake_ssh_exec(*args, **kwargs):
            if 'showshare' in args[1]:
                raise exception.ProcessExecutionError()
            else:
                return '', ''

        self.mock_object(self._helper, '_ssh_exec',
                         mock.Mock(side_effect=fake_ssh_exec))

        ret = self._helper.create_exports(self.server_details, self.share_name)

        expected_location = [{
            "is_admin_only": False,
            "path": "\\\\%s\\%s" % (
                self.server_details['public_address'], self.share_name),
            "metadata": {"export_location_metadata_example": "example"}
        }]
        self.assertEqual(expected_location, ret)
        share_path = os.path.join(
            self._helper.configuration.share_mount_path,
            self.share_name)
        self._helper._ssh_exec.assert_has_calls([
            mock.call(
                self.server_details,
                ['sudo', 'net', 'conf', 'showshare', self.share_name, ]
            ),
            mock.call(
                self.server_details,
                [
                    'sudo', 'net', 'conf', 'addshare', self.share_name,
                    share_path, 'writeable=y', 'guest_ok=y',
                ]
            ),
            mock.call(self.server_details, mock.ANY),
        ])

    def test_create_export_share_does_not_exist_exception(self):

        self.mock_object(self._helper, '_ssh_exec',
                         mock.Mock(
                             side_effect=[exception.ProcessExecutionError(),
                                          Exception('')]
                         ))

        self.assertRaises(
            exception.ManilaException, self._helper.create_exports,
            self.server_details, self.share_name)

    def test_create_exports_share_exist_recreate_true(self):
        ret = self._helper.create_exports(
            self.server_details, self.share_name, recreate=True)

        expected_location = [{
            "is_admin_only": False,
            "path": "\\\\%s\\%s" % (
                self.server_details['public_address'], self.share_name),
            "metadata": {"export_location_metadata_example": "example"}
        }]
        self.assertEqual(expected_location, ret)
        share_path = os.path.join(
            self._helper.configuration.share_mount_path,
            self.share_name)
        self._helper._ssh_exec.assert_has_calls([
            mock.call(
                self.server_details,
                ['sudo', 'net', 'conf', 'showshare', self.share_name, ]
            ),
            mock.call(
                self.server_details,
                ['sudo', 'net', 'conf', 'delshare', self.share_name, ]
            ),
            mock.call(
                self.server_details,
                [
                    'sudo', 'net', 'conf', 'addshare', self.share_name,
                    share_path, 'writeable=y', 'guest_ok=y',
                ]
            ),
            mock.call(self.server_details, mock.ANY),
        ])

    def test_create_export_share_exist_recreate_false(self):
        self.assertRaises(
            exception.ShareBackendException,
            self._helper.create_exports,
            self.server_details,
            self.share_name,
            recreate=False,
        )
        self._helper._ssh_exec.assert_has_calls([
            mock.call(
                self.server_details,
                ['sudo', 'net', 'conf', 'showshare', self.share_name, ]
            ),
        ])

    def test_remove_exports(self):
        self._helper.remove_exports(self.server_details, self.share_name)

        self._helper._ssh_exec.assert_called_once_with(
            self.server_details,
            ['sudo', 'net', 'conf', 'delshare', self.share_name],
        )

    def test_remove_export_forcibly(self):
        delshare_command = ['sudo', 'net', 'conf', 'delshare', self.share_name]

        def fake_ssh_exec(*args, **kwargs):
            if delshare_command == args[1]:
                raise exception.ProcessExecutionError()
            else:
                return ('', '')

        self.mock_object(self._helper, '_ssh_exec',
                         mock.Mock(side_effect=fake_ssh_exec))

        self._helper.remove_exports(self.server_details, self.share_name)

        self._helper._ssh_exec.assert_has_calls([
            mock.call(
                self.server_details,
                ['sudo', 'net', 'conf', 'delshare', self.share_name],
            ),
            mock.call(
                self.server_details,
                ['sudo', 'smbcontrol', 'all', 'close-share', self.share_name],
            ),
        ])

    def test_update_access_wrong_access_level(self):
        access_rules = [test_generic.get_fake_access_rule(
            '2.2.2.2', const.ACCESS_LEVEL_RO), ]
        self.assertRaises(
            exception.InvalidShareAccessLevel,
            self._helper.update_access,
            self.server_details,
            self.share_name,
            access_rules,
            [],
            [])

    def test_update_access_wrong_access_type(self):
        access_rules = [test_generic.get_fake_access_rule(
            '2.2.2.2', const.ACCESS_LEVEL_RW, access_type='fake'), ]
        self.assertRaises(
            exception.InvalidShareAccess,
            self._helper.update_access,
            self.server_details,
            self.share_name,
            access_rules,
            [],
            [])

    def test_update_access(self):
        access_rules = [test_generic.get_fake_access_rule(
            '1.1.1.1', const.ACCESS_LEVEL_RW), ]

        self._helper.update_access(self.server_details, self.share_name,
                                   access_rules, [], [])
        self._helper._ssh_exec.assert_called_once_with(
            self.server_details, ['sudo', 'net', 'conf', 'setparm',
                                  self.share_name, 'hosts allow',
                                  '1.1.1.1'])

    def test_get_allow_hosts(self):
        self.mock_object(self._helper, '_ssh_exec',
                         mock.Mock(
                             return_value=('1.1.1.1 2.2.2.2 3.3.3.3', '')))
        expected = ['1.1.1.1', '2.2.2.2', '3.3.3.3']
        result = self._helper._get_allow_hosts(
            self.server_details, self.share_name)
        self.assertEqual(expected, result)
        cmd = ['sudo', 'net', 'conf', 'getparm', self.share_name,
               'hosts allow']
        self._helper._ssh_exec.assert_called_once_with(
            self.server_details, cmd)

    @ddt.data(
        '', '1.2.3.4:/nfs/like/export', '/1.2.3.4/foo', '\\1.2.3.4\\foo',
        '//1.2.3.4\\mixed_slashes_and_backslashes_one',
        '\\\\1.2.3.4/mixed_slashes_and_backslashes_two')
    def test__get_share_group_name_from_export_location(self, export_location):
        self.assertRaises(
            exception.InvalidShare,
            self._helper._get_share_group_name_from_export_location,
            export_location)

    @ddt.data('//5.6.7.8/foo', '\\\\5.6.7.8\\foo')
    def test_get_exports_for_share(self, export_location):
        server = dict(public_address='1.2.3.4')
        self.mock_object(
            self._helper, '_get_share_group_name_from_export_location',
            mock.Mock(side_effect=(
                self._helper._get_share_group_name_from_export_location)))

        result = self._helper.get_exports_for_share(server, export_location)

        expected_export_location = [{
            "is_admin_only": False,
            "path": "\\\\%s\\foo" % server['public_address'],
            "metadata": {"export_location_metadata_example": "example"}
        }]
        self.assertEqual(expected_export_location, result)
        (self._helper._get_share_group_name_from_export_location.
            assert_called_once_with(export_location))

    @ddt.data(
        {'public_address_with_suffix': 'foo'},
        {'with_prefix_public_address': 'bar'},
        {'with_prefix_public_address_and_with_suffix': 'quuz'}, {})
    def test_get_exports_for_share_with_exception(self, server):
        export_location = '1.2.3.4:/foo/bar'

        self.assertRaises(
            exception.ManilaException,
            self._helper.get_exports_for_share, server, export_location)

    @ddt.data('//5.6.7.8/foo', '\\\\5.6.7.8\\foo')
    def test_get_share_path_by_export_location(self, export_location):
        fake_path = ' /bar/quuz\n '
        fake_server = dict()
        self.mock_object(
            self._helper, '_ssh_exec',
            mock.Mock(return_value=(fake_path, 'fake')))
        self.mock_object(
            self._helper, '_get_share_group_name_from_export_location',
            mock.Mock(side_effect=(
                self._helper._get_share_group_name_from_export_location)))

        result = self._helper.get_share_path_by_export_location(
            fake_server, export_location)

        self.assertEqual('/bar/quuz', result)
        self._helper._ssh_exec.assert_called_once_with(
            fake_server, ['sudo', 'net', 'conf', 'getparm', 'foo', 'path'])
        (self._helper._get_share_group_name_from_export_location.
            assert_called_once_with(export_location))

    def test_disable_access_for_maintenance(self):
        allowed_hosts = ['test', 'test2']
        maintenance_path = os.path.join(
            self._helper.configuration.share_mount_path,
            "%s.maintenance" % self.share_name)
        self.mock_object(self._helper, '_set_allow_hosts')
        self.mock_object(self._helper, '_get_allow_hosts',
                         mock.Mock(return_value=allowed_hosts))

        self._helper.disable_access_for_maintenance(
            self.server_details, self.share_name)

        self._helper._get_allow_hosts.assert_called_once_with(
            self.server_details, self.share_name)
        self._helper._set_allow_hosts.assert_called_once_with(
            self.server_details, [], self.share_name)
        kickoff_user_cmd = ['sudo', 'smbstatus', '-S']
        self._helper._ssh_exec.assert_any_call(
            self.server_details, kickoff_user_cmd)
        valid_cmd = ['echo', "'test test2'", '|', 'sudo', 'tee',
                     maintenance_path]
        self._helper._ssh_exec.assert_any_call(
            self.server_details, valid_cmd)

    def test__kick_out_users_success(self):
        smbstatus_return = """Service      pid     machine       Connected at
-------------------------------------------------------
fake_share_name   1001   fake_machine1    Thu Sep 14 14:59:07 2017
fake_share_name   1002   fake_machine2    Thu Sep 14 14:59:07 2017
"""
        self.mock_object(self._helper, '_ssh_exec', mock.Mock(
            side_effect=[(smbstatus_return, "fake_stderr"), ("fake", "fake")]))
        self._helper._kick_out_users(self.server_details, self.share_name)
        self._helper._ssh_exec.assert_any_call(
            self.server_details, ['sudo', 'smbstatus', '-S'])
        self._helper._ssh_exec.assert_any_call(
            self.server_details, ["sudo", "kill", "-15", "1001", "1002"])

    def test__kick_out_users_failed(self):
        smbstatus_return = """Service      pid     machine       Connected at
-------------------------------------------------------
fake line
"""
        self.mock_object(self._helper, '_ssh_exec', mock.Mock(
            return_value=(smbstatus_return, "fake_stderr")))
        self.assertRaises(exception.ShareBackendException,
                          self._helper._kick_out_users, self.server_details,
                          self.share_name)

    def test_restore_access_after_maintenance(self):
        fake_maintenance_path = "test.path"
        self.mock_object(self._helper, '_set_allow_hosts')
        self.mock_object(self._helper, '_get_maintenance_file_path',
                         mock.Mock(return_value=fake_maintenance_path))
        self.mock_object(self._helper, '_ssh_exec',
                         mock.Mock(side_effect=[("fake fake2", 0), "fake"]))

        self._helper.restore_access_after_maintenance(
            self.server_details, self.share_name)

        self._helper._set_allow_hosts.assert_called_once_with(
            self.server_details, ['fake', 'fake2'], self.share_name)
        self._helper._ssh_exec.assert_any_call(
            self.server_details, ['cat', fake_maintenance_path])
        self._helper._ssh_exec.assert_any_call(
            self.server_details, ['sudo', 'rm', '-f', fake_maintenance_path])


@ddt.ddt
class CIFSHelperUserAccessTestCase(test.TestCase):
    """Test case for CIFS helper with user access."""
    access_rw = dict(
        access_level=const.ACCESS_LEVEL_RW,
        access_type='user',
        access_to='manila-user')
    access_ro = dict(
        access_level=const.ACCESS_LEVEL_RO,
        access_type='user',
        access_to='manila-user')

    def setUp(self):
        super(CIFSHelperUserAccessTestCase, self).setUp()
        self.server_details = {'instance_id': 'fake',
                               'public_address': '1.2.3.4', }
        self.share_name = 'fake_share_name'
        self.fake_conf = manila.share.configuration.Configuration(None)
        self._ssh_exec = mock.Mock(return_value=('', ''))
        self._execute = mock.Mock(return_value=('', ''))
        self._helper = helpers.CIFSHelperUserAccess(
            self._execute, self._ssh_exec, self.fake_conf)

    def test_update_access_exception_type(self):
        access_rules = [test_generic.get_fake_access_rule(
            'user1', const.ACCESS_LEVEL_RW, access_type='ip')]
        self.assertRaises(exception.InvalidShareAccess,
                          self._helper.update_access, self.server_details,
                          self.share_name, access_rules, [], [])

    def test_update_access(self):
        access_list = [test_generic.get_fake_access_rule(
            'user1', const.ACCESS_LEVEL_RW, access_type='user'),
            test_generic.get_fake_access_rule(
                'user2', const.ACCESS_LEVEL_RO, access_type='user')]
        self._helper.update_access(self.server_details, self.share_name,
                                   access_list, [], [])

        self._helper._ssh_exec.assert_has_calls([
            mock.call(self.server_details,
                      ['sudo', 'net', 'conf', 'setparm', self.share_name,
                       'valid users', 'user1']),
            mock.call(self.server_details,
                      ['sudo', 'net', 'conf', 'setparm', self.share_name,
                       'read list', 'user2'])
        ])

    def test_update_access_exception_level(self):
        access_rules = [test_generic.get_fake_access_rule(
            'user1', 'fake_level', access_type='user'), ]
        self.assertRaises(
            exception.InvalidShareAccessLevel,
            self._helper.update_access,
            self.server_details,
            self.share_name,
            access_rules,
            [],
            [])


@ddt.ddt
class NFSSynchronizedTestCase(test.TestCase):

    @helpers.nfs_synchronized
    def wrapped_method(self, server, share_name):
        return server['instance_id'] + share_name

    @ddt.data(
        ({'lock_name': 'FOO', 'instance_id': 'QUUZ'}, 'nfs-FOO'),
        ({'instance_id': 'QUUZ'}, 'nfs-QUUZ'),
    )
    @ddt.unpack
    def test_with_lock_name(self, server, expected_lock_name):
        share_name = 'fake_share_name'
        self.mock_object(
            helpers.utils, 'synchronized',
            mock.Mock(side_effect=helpers.utils.synchronized))

        result = self.wrapped_method(server, share_name)

        self.assertEqual(server['instance_id'] + share_name, result)
        helpers.utils.synchronized.assert_called_once_with(
            expected_lock_name, external=True)



# Copyright (c) 2020 Zadara Storage, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import re

from oslo_config import cfg
from oslo_log import log as logging
import requests

LOG = logging.getLogger(__name__)

# Number of seconds the repsonse for the request sent to
# vpsa is expected. Else the request will be timed out.
# Setting it to 300 seconds initially.
vpsa_timeout = 300


# Common exception class for all the exceptions that
# are used to redirect to the driver specific exceptions.
class CommonException(Exception):
    def __init__(self):
        pass

    class UnknownCmd(Exception):
        def __init__(self, cmd):
            self.cmd = cmd

    class BadHTTPResponseStatus(Exception):
        def __init__(self, status):
            self.status = status

    class FailedCmdWithDump(Exception):
        def __init__(self, status, data):
            self.status = status
            self.data = data

    class SessionRequestException(Exception):
        def __init__(self, msg):
            self.msg = msg

    class ZadaraInvalidAccessKey(Exception):
        pass


exception = CommonException()


zadara_opts = [
    cfg.HostAddressOpt('zadara_vpsa_host',
                       default=None,
                       help='VPSA - Management Host name or IP address'),
    cfg.PortOpt('zadara_vpsa_port',
                default=None,
                help='VPSA - Port number'),
    cfg.BoolOpt('zadara_vpsa_use_ssl',
                default=False,
                help='VPSA - Use SSL connection'),
    cfg.BoolOpt('zadara_ssl_cert_verify',
                default=True,
                help='If set to True the http client will validate the SSL '
                     'certificate of the VPSA endpoint.'),
    cfg.StrOpt('zadara_access_key',
               default=None,
               help='VPSA access key',
               secret=True),
    cfg.StrOpt('zadara_vpsa_poolname',
               default=None,
               help='VPSA - Storage Pool assigned for volumes'),
    cfg.BoolOpt('zadara_vol_encrypt',
                default=False,
                help='VPSA - Default encryption policy for volumes. '
                     'If the option is neither configured nor provided '
                     'as metadata, the VPSA will inherit the default value.'),
    cfg.BoolOpt('zadara_gen3_vol_dedupe',
                default=False,
                help='VPSA - Enable deduplication for volumes. '
                     'If the option is neither configured nor provided '
                     'as metadata, the VPSA will inherit the default value.'),
    cfg.BoolOpt('zadara_gen3_vol_compress',
                default=False,
                help='VPSA - Enable compression for volumes. '
                     'If the option is neither configured nor provided '
                     'as metadata, the VPSA will inherit the default value.'),
    cfg.BoolOpt('zadara_default_snap_policy',
                default=False,
                help="VPSA - Attach snapshot policy for volumes. "
                     "If the option is neither configured nor provided "
                     "as metadata, the VPSA will inherit the default value.")]


# Class used to connect and execute the commands on
# Zadara Virtual Private Storage Array (VPSA).
class ZadaraVPSAConnection(object):
    """Executes driver commands on VPSA."""

    def __init__(self, conf, driver_ssl_cert_path, block):
        self.conf = conf
        self.access_key = conf.zadara_access_key
        if not self.access_key:
            raise exception.ZadaraInvalidAccessKey()
        self.driver_ssl_cert_path = driver_ssl_cert_path
        # Choose the volume type of either block or file-type
        # that will help to filter volumes.
        self.vol_type_str = 'showonlyblock' if block else 'showonlyfile'

    def _generate_vpsa_cmd(self, cmd, **kwargs):
        """Generate command to be sent to VPSA."""

        # Dictionary of applicable VPSA commands in the following format:
        # 'command': (method, API_URL, {optional parameters})
        vpsa_commands = {
            # Volume operations
            'create_volume': ('POST',
                              '/api/volumes.json',
                              {'name': kwargs.get('name'),
                               'capacity': kwargs.get('size'),
                               'pool': self.conf.zadara_vpsa_poolname,
                               'block': 'YES'
                               if self.vol_type_str == 'showonlyblock'
                               else 'NO',
                               'thin': 'YES',
                               'crypt': 'YES'
                               if self.conf.zadara_vol_encrypt else 'NO',
                               'compress': 'YES'
                               if self.conf.zadara_gen3_vol_compress else 'NO',
                               'dedupe': 'YES'
                               if self.conf.zadara_gen3_vol_dedupe else 'NO',
                               'attachpolicies': 'NO'
                               if not self.conf.zadara_default_snap_policy
                               else 'YES'}),
            'delete_volume': ('DELETE',
                              '/api/volumes/%s.json' % kwargs.get('vpsa_vol'),
                              {'force': 'YES'}),
            'expand_volume': ('POST',
                              '/api/volumes/%s/expand.json'
                              % kwargs.get('vpsa_vol'),
                              {'capacity': kwargs.get('size')}),
            'rename_volume': ('POST',
                              '/api/volumes/%s/rename.json'
                              % kwargs.get('vpsa_vol'),
                              {'new_name': kwargs.get('new_name')}),
            # Snapshot operations
            # Snapshot request is triggered for a single volume though the
            # API call implies that snapshot is triggered for CG (legacy API).
            'create_snapshot': ('POST',
                                '/api/consistency_groups/%s/snapshots.json'
                                % kwargs.get('cg_name'),
                                {'display_name': kwargs.get('snap_name')}),
            'delete_snapshot': ('DELETE',
                                '/api/snapshots/%s.json'
                                % kwargs.get('snap_id'),
                                {}),
            'rename_snapshot': ('POST',
                                '/api/snapshots/%s/rename.json'
                                % kwargs.get('snap_id'),
                                {'newname': kwargs.get('new_name')}),
            'create_clone_from_snap': ('POST',
                                       '/api/consistency_groups/%s/clone.json'
                                       % kwargs.get('cg_name'),
                                       {'name': kwargs.get('name'),
                                        'snapshot': kwargs.get('snap_id')}),
            'create_clone': ('POST',
                             '/api/consistency_groups/%s/clone.json'
                             % kwargs.get('cg_name'),
                             {'name': kwargs.get('name')}),
            # Server operations
            'create_server': ('POST',
                              '/api/servers.json',
                              {'iqn': kwargs.get('iqn'),
                               'iscsi': kwargs.get('iscsi_ip'),
                               'display_name': kwargs.get('iqn')
                               if kwargs.get('iqn')
                               else kwargs.get('iscsi_ip')}),
            # Attach/Detach operations
            'attach_volume': ('POST',
                              '/api/servers/%s/volumes.json'
                              % kwargs.get('vpsa_srv'),
                              {'volume_name[]': kwargs.get('vpsa_vol'),
                               'access_type': kwargs.get('share_proto'),
                               'readonly': kwargs.get('read_only'),
                               'force': 'YES'}),
            'detach_volume': ('POST',
                              '/api/volumes/%s/detach.json'
                              % kwargs.get('vpsa_vol'),
                              {'server_name[]': kwargs.get('vpsa_srv'),
                               'force': 'YES'}),
            # Update volume comment
            'update_volume': ('POST',
                              '/api/volumes/%s/update_comment.json'
                              % kwargs.get('vpsa_vol'),
                              {'new_comment': kwargs.get('new_comment')}),

            # Get operations
            'list_volumes': ('GET',
                             '/api/volumes.json?%s=YES' % self.vol_type_str,
                             {}),
            'get_volume': ('GET',
                           '/api/volumes/%s.json' % kwargs.get('vpsa_vol'),
                           {}),
            'get_volume_by_name': ('GET',
                                   '/api/volumes.json?display_name=%s'
                                   % kwargs.get('display_name'),
                                   {}),
            'get_pool': ('GET',
                         '/api/pools/%s.json' % kwargs.get('pool_name'),
                         {}),
            'list_controllers': ('GET',
                                 '/api/vcontrollers.json',
                                 {}),
            'list_servers': ('GET',
                             '/api/servers.json',
                             {}),
            'list_vol_snapshots': ('GET',
                                   '/api/consistency_groups/%s/snapshots.json'
                                   % kwargs.get('cg_name'),
                                   {}),
            'list_vol_attachments': ('GET',
                                     '/api/volumes/%s/servers.json'
                                     % kwargs.get('vpsa_vol'),
                                     {}),
            'list_snapshots': ('GET',
                               '/api/snapshots.json',
                               {}),
            # Put operations
            'change_export_name': ('PUT',
                                   '/api/volumes/%s/export_name.json'
                                   % kwargs.get('vpsa_vol'),
                                   {'exportname': kwargs.get('exportname')})}
        try:
            method, url, params = vpsa_commands[cmd]
            # Populate the metadata for the volume creation
            metadata = kwargs.get('metadata')
            if metadata:
                for key, value in metadata.items():
                    params[key] = value
        except KeyError:
            raise exception.UnknownCmd(cmd=cmd)

        if method == 'GET':
            params = dict(page=1, start=0, limit=0)
            body = None

        elif method in ['DELETE', 'POST', 'PUT']:
            body = params
            params = None

        else:
            msg = ('Method %(method)s is not defined' % {'method': method})
            LOG.error(msg)
            raise AssertionError(msg)

        # 'access_key' was generated using username and password
        # or it was taken from the input file
        headers = {'X-Access-Key': self.access_key}

        return method, url, params, body, headers

    def send_cmd(self, cmd, **kwargs):
        """Send command to VPSA Controller."""

        if not self.access_key:
            raise exception.ZadaraInvalidAccessKey()

        method, url, params, body, headers = self._generate_vpsa_cmd(cmd,
                                                                     **kwargs)
        LOG.debug('Invoking %(cmd)s using %(method)s request.',
                  {'cmd': cmd, 'method': method})

        host = self._get_target_host(self.conf.zadara_vpsa_host)
        port = int(self.conf.zadara_vpsa_port)

        protocol = "https" if self.conf.zadara_vpsa_use_ssl else "http"
        if protocol == "https":
            if not self.conf.zadara_ssl_cert_verify:
                verify = False
            else:
                verify = (self.driver_ssl_cert_path
                          if self.driver_ssl_cert_path else True)
        else:
            verify = False

        if port:
            api_url = "%s://%s:%d%s" % (protocol, host, port, url)
        else:
            api_url = "%s://%s%s" % (protocol, host, url)

        try:
            with requests.Session() as session:
                session.headers.update(headers)
                response = session.request(method, api_url, params=params,
                                           data=body, headers=headers,
                                           verify=verify, timeout=vpsa_timeout)
        except requests.exceptions.RequestException as e:
            msg = ('Exception: %s') % e
            raise exception.SessionRequestException(msg=msg)

        if response.status_code != 200:
            raise exception.BadHTTPResponseStatus(
                status=response.status_code)

        data = response.content
        json_data = json.loads(data)
        response = json_data['response']
        status = int(response['status'])
        if status == 5:
            # Invalid Credentials
            raise exception.ZadaraInvalidAccessKey()

        if status != 0:
            raise exception.FailedCmdWithDump(status=status, data=data)

        if method in ['POST', 'DELETE']:
            LOG.debug('Operation completed with status code %(status)s',
                      {'status': status})
        return response

    def _get_target_host(self, vpsa_host):
        """Helper for target host formatting."""
        ipv6_without_brackets = ':' in vpsa_host and vpsa_host[-1] != ']'
        if ipv6_without_brackets:
            return ('[%s]' % vpsa_host)
        return ('%s' % vpsa_host)

    def _get_active_controller_details(self):
        """Return details of VPSA's active controller."""
        data = self.send_cmd('list_controllers')
        ctrl = None
        vcontrollers = data.get('vcontrollers', [])
        for controller in vcontrollers:
            if controller['state'] == 'active':
                ctrl = controller
                break

        if ctrl is not None:
            target_ip = (ctrl['iscsi_ipv6'] if
                         ctrl['iscsi_ipv6'] else
                         ctrl['iscsi_ip'])
            return dict(target=ctrl['target'],
                        ip=target_ip,
                        chap_user=ctrl['vpsa_chap_user'],
                        chap_passwd=ctrl['vpsa_chap_secret'])
        return None

    def _check_access_key_validity(self):
        """Check VPSA access key"""
        if not self.access_key:
            raise exception.ZadaraInvalidAccessKey()
        active_ctrl = self._get_active_controller_details()
        if active_ctrl is None:
            raise exception.ZadaraInvalidAccessKey()

    def _get_vpsa_volume(self, name):
        """Returns a single vpsa volume based on the display name"""
        volume = None
        display_name = name
        if re.search(r"\s", name):
            display_name = re.split(r"\s", name)[0]
        data = self.send_cmd('get_volume_by_name',
                             display_name=display_name)
        if data['status'] != 0:
            return None
        volumes = data['volumes']

        for vol in volumes:
            if vol['display_name'] == name:
                volume = vol
                break
        return volume

    def _get_vpsa_volume_by_id(self, vpsa_vol):
        """Returns a single vpsa volume based on name"""
        data = self.send_cmd('get_volume', vpsa_vol=vpsa_vol)
        return data['volume']

    def _get_volume_cg_name(self, name):
        """Return name of the consistency group for the volume.

        cg-name is a volume uniqe identifier (legacy attribute)
        and not consistency group as it may imply.
        """
        volume = self._get_vpsa_volume(name)
        if volume is not None:
            return volume['cg_name']

        return None

    def _get_all_vpsa_snapshots(self):
        """Returns snapshots from all vpsa volumes"""
        data = self.send_cmd('list_snapshots')
        return data['snapshots']

    def _get_all_vpsa_volumes(self):
        """Returns all vpsa block volumes from the configured pool"""
        data = self.send_cmd('list_volumes')
        # FIXME: Work around to filter volumes belonging to given pool
        # Remove this when we have the API fixed to filter based
        # on pools. This API today does not have virtual_capacity field
        volumes = []

        for volume in data['volumes']:
            if volume['pool_name'] == self.conf.zadara_vpsa_poolname:
                volumes.append(volume)

        return volumes

    def _get_server_name(self, initiator, share):
        """Return VPSA's name for server object.

           'share' will be true to search for filesystem volumes
        """
        data = self.send_cmd('list_servers')
        servers = data.get('servers', [])
        for server in servers:
            if share:
                if server['iscsi_ip'] == initiator:
                    return server['name']
            else:
                if server['iqn'] == initiator:
                    return server['name']
        return None

    def _create_vpsa_server(self, iqn=None, iscsi_ip=None):
        """Create server object within VPSA (if doesn't exist)."""
        initiator = iscsi_ip if iscsi_ip else iqn
        share = True if iscsi_ip else False
        vpsa_srv = self._get_server_name(initiator, share)
        if not vpsa_srv:
            data = self.send_cmd('create_server', iqn=iqn, iscsi_ip=iscsi_ip)
            if data['status'] != 0:
                return None
            vpsa_srv = data['server_name']
        return vpsa_srv

    def _get_servers_attached_to_volume(self, vpsa_vol):
        """Return all servers attached to volume."""
        servers = vpsa_vol.get('server_ext_names')
        list_servers = []
        if servers:
            list_servers = servers.split(',')
        return list_servers

    def _detach_vpsa_volume(self, vpsa_vol, vpsa_srv=None):
        """Detach volume from all attached servers."""
        if vpsa_srv:
            list_servers_ids = [vpsa_srv]
        else:
            list_servers_ids = self._get_servers_attached_to_volume(vpsa_vol)

        for server_id in list_servers_ids:
            # Detach volume from server
            self.send_cmd('detach_volume', vpsa_srv=server_id,
                          vpsa_vol=vpsa_vol['name'])

    def _get_volume_snapshots(self, cg_name):
        """Get snapshots in the consistency group"""
        data = self.send_cmd('list_vol_snapshots', cg_name=cg_name)
        snapshots = data.get('snapshots', [])
        return snapshots

    def _get_snap_id(self, cg_name, snap_name):
        """Return snapshot ID for particular volume."""
        snapshots = self._get_volume_snapshots(cg_name)
        for snap_vol in snapshots:
            if snap_vol['display_name'] == snap_name:
                return snap_vol['name']

        return None

    def _get_pool_capacity(self, pool_name):
        """Return pool's total and available capacities."""
        data = self.send_cmd('get_pool', pool_name=pool_name)
        pool = data.get('pool')
        if pool is not None:
            total = int(pool['capacity'])
            free = int(pool['available_capacity'])
            provisioned = int(pool['provisioned_capacity'])
            LOG.debug('Pool %(name)s: %(total)sGB total, %(free)sGB free, '
                      '%(provisioned)sGB provisioned',
                      {'name': pool_name, 'total': total,
                       'free': free, 'provisioned': provisioned})
            return total, free, provisioned

        return 'unknown', 'unknown', 'unknown'



# Copyright (c) 2015 Hitachi Data Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from oslo_config import cfg
from oslo_log import log
from oslo_utils import excutils
from oslo_utils import importutils
import six

from manila.common import constants
from manila import exception
from manila.i18n import _
from manila.share import driver
from manila.share import utils

LOG = log.getLogger(__name__)

hitachi_hnas_opts = [
    cfg.HostAddressOpt('hitachi_hnas_ip',
                       help="HNAS management interface IP for communication "
                            "between Manila controller and HNAS."),
    cfg.StrOpt('hitachi_hnas_user',
               help="HNAS username Base64 String in order to perform tasks "
                    "such as create file-systems and network interfaces."),
    cfg.StrOpt('hitachi_hnas_password',
               secret=True,
               help="HNAS user password. Required only if private key is not "
                    "provided."),
    cfg.IntOpt('hitachi_hnas_evs_id',
               help="Specify which EVS this backend is assigned to."),
    cfg.HostAddressOpt('hitachi_hnas_evs_ip',
                       help="Specify IP for mounting shares."),
    cfg.HostAddressOpt('hitachi_hnas_admin_network_ip',
                       help="Specify IP for mounting shares in the Admin "
                            "network."),
    cfg.StrOpt('hitachi_hnas_file_system_name',
               help="Specify file-system name for creating shares."),
    cfg.StrOpt('hitachi_hnas_ssh_private_key',
               secret=True,
               help="RSA/DSA private key value used to connect into HNAS. "
                    "Required only if password is not provided."),
    cfg.HostAddressOpt('hitachi_hnas_cluster_admin_ip0',
                       help="The IP of the clusters admin node. Only set in "
                            "HNAS multinode clusters."),
    cfg.IntOpt('hitachi_hnas_stalled_job_timeout',
               default=30,
               help="The time (in seconds) to wait for stalled HNAS jobs "
                    "before aborting."),
    cfg.StrOpt('hitachi_hnas_driver_helper',
               default='manila.share.drivers.hitachi.hnas.ssh.HNASSSHBackend',
               help="Python class to be used for driver helper."),
    cfg.BoolOpt('hitachi_hnas_allow_cifs_snapshot_while_mounted',
                default=False,
                help="By default, CIFS snapshots are not allowed to be taken "
                     "when the share has clients connected because consistent "
                     "point-in-time replica cannot be guaranteed for all "
                     "files. Enabling this might cause inconsistent snapshots "
                     "on CIFS shares."),
]

CONF = cfg.CONF
CONF.register_opts(hitachi_hnas_opts)


class HitachiHNASDriver(driver.ShareDriver):
    """Manila HNAS Driver implementation.

    Driver versions::

        1.0.0 - Initial Version.
        2.0.0 - Refactoring, bugfixes, implemented Share Shrink and
                Update Access.
        3.0.0 - New driver location, implemented support for CIFS protocol.
        3.1.0 - Added admin network export location support.
        4.0.0 - Added mountable snapshots, revert-to-snapshot and
                manage snapshots features support.
    """

    def __init__(self, *args, **kwargs):
        """Do initialization."""

        LOG.debug("Invoking base constructor for Manila Hitachi HNAS Driver.")
        super(HitachiHNASDriver, self).__init__(False, *args, **kwargs)

        LOG.debug("Setting up attributes for Manila Hitachi HNAS Driver.")
        self.configuration.append_config_values(hitachi_hnas_opts)

        LOG.debug("Reading config parameters for Manila Hitachi HNAS Driver.")
        self.backend_name = self.configuration.safe_get('share_backend_name')
        hnas_helper = self.configuration.safe_get('hitachi_hnas_driver_helper')
        hnas_ip = self.configuration.safe_get('hitachi_hnas_ip')
        hnas_username = self.configuration.safe_get('hitachi_hnas_user')
        hnas_password = self.configuration.safe_get('hitachi_hnas_password')
        hnas_evs_id = self.configuration.safe_get('hitachi_hnas_evs_id')
        self.hnas_evs_ip = self.configuration.safe_get('hitachi_hnas_evs_ip')
        self.hnas_admin_network_ip = self.configuration.safe_get(
            'hitachi_hnas_admin_network_ip')
        self.fs_name = self.configuration.safe_get(
            'hitachi_hnas_file_system_name')
        self.cifs_snapshot = self.configuration.safe_get(
            'hitachi_hnas_allow_cifs_snapshot_while_mounted')
        ssh_private_key = self.configuration.safe_get(
            'hitachi_hnas_ssh_private_key')
        cluster_admin_ip0 = self.configuration.safe_get(
            'hitachi_hnas_cluster_admin_ip0')
        self.private_storage = kwargs.get('private_storage')
        job_timeout = self.configuration.safe_get(
            'hitachi_hnas_stalled_job_timeout')

        if hnas_helper is None:
            msg = _("The config parameter hitachi_hnas_driver_helper is not "
                    "set.")
            raise exception.InvalidParameterValue(err=msg)

        if hnas_evs_id is None:
            msg = _("The config parameter hitachi_hnas_evs_id is not set.")
            raise exception.InvalidParameterValue(err=msg)

        if self.hnas_evs_ip is None:
            msg = _("The config parameter hitachi_hnas_evs_ip is not set.")
            raise exception.InvalidParameterValue(err=msg)

        if hnas_ip is None:
            msg = _("The config parameter hitachi_hnas_ip is not set.")
            raise exception.InvalidParameterValue(err=msg)

        if hnas_username is None:
            msg = _("The config parameter hitachi_hnas_user is not set.")
            raise exception.InvalidParameterValue(err=msg)

        if hnas_password is None and ssh_private_key is None:
            msg = _("Credentials configuration parameters missing: "
                    "you need to set hitachi_hnas_password or "
                    "hitachi_hnas_ssh_private_key.")
            raise exception.InvalidParameterValue(err=msg)

        LOG.debug("Initializing HNAS Layer.")

        helper = importutils.import_class(hnas_helper)

        self.hnas = helper(hnas_ip, hnas_username, hnas_password,
                           ssh_private_key, cluster_admin_ip0,
                           hnas_evs_id, self.hnas_evs_ip, self.fs_name,
                           job_timeout)

    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        """Update access rules for given share.

        :param context: The `context.RequestContext` object for the request
        :param share: Share that will have its access rules updated.
        :param access_rules: All access rules for given share.
        :param add_rules: Empty List or List of access rules which should be
            added. access_rules already contains these rules.
        :param delete_rules: Empty List or List of access rules which should be
            removed. access_rules doesn't contain these rules.
        :param share_server: Data structure with share server information.
            Not used by this driver.
        """

        hnas_share_id = self._get_hnas_share_id(share['id'])

        try:
            self._ensure_share(share, hnas_share_id)
        except exception.HNASItemNotFoundException:
            raise exception.ShareResourceNotFound(share_id=share['id'])

        self._check_protocol(share['id'], share['share_proto'])

        if share['share_proto'].lower() == 'nfs':
            self._nfs_update_access(share, hnas_share_id, access_rules)
        else:
            if not (add_rules or delete_rules):
                # recovery mode
                self._clean_cifs_access_list(hnas_share_id)
                self._cifs_allow_access(share, hnas_share_id, access_rules)
            else:
                self._cifs_deny_access(share, hnas_share_id, delete_rules)
                self._cifs_allow_access(share, hnas_share_id, add_rules)

    def _nfs_update_access(self, share, hnas_share_id, access_rules):
        host_list = []

        for rule in access_rules:
            if rule['access_type'].lower() != 'ip':
                msg = _("Only IP access type currently supported for NFS. "
                        "Share provided %(share)s with rule type "
                        "%(type)s.") % {'share': share['id'],
                                        'type': rule['access_type']}
                raise exception.InvalidShareAccess(reason=msg)

            if rule['access_level'] == constants.ACCESS_LEVEL_RW:
                host_list.append(rule['access_to'] + '(' +
                                 rule['access_level'] +
                                 ',norootsquash)')
            else:
                host_list.append(rule['access_to'] + '(' +
                                 rule['access_level'] + ')')

        self.hnas.update_nfs_access_rule(host_list, share_id=hnas_share_id)

        if host_list:
            LOG.debug("Share %(share)s has the rules: %(rules)s",
                      {'share': share['id'], 'rules': ', '.join(host_list)})
        else:
            LOG.debug("Share %(share)s has no rules.", {'share': share['id']})

    def _cifs_allow_access(self, share_or_snapshot, hnas_id, add_rules,
                           is_snapshot=False):

        entity_type = "share"
        if is_snapshot:
            entity_type = "snapshot"

        for rule in add_rules:
            if rule['access_type'].lower() != 'user':
                msg = _("Only USER access type currently supported for CIFS. "
                        "%(entity_type)s provided %(share)s with "
                        "rule %(r_id)s type %(type)s allowing permission "
                        "to %(to)s.") % {
                    'entity_type': entity_type.capitalize(),
                    'share': share_or_snapshot['id'],
                    'type': rule['access_type'],
                    'r_id': rule['id'],
                    'to': rule['access_to'],
                }
                raise exception.InvalidShareAccess(reason=msg)

            if rule['access_level'] == constants.ACCESS_LEVEL_RW:
                # Adding permission acr = Allow Change&Read
                permission = 'acr'
            else:
                # Adding permission ar = Allow Read
                permission = 'ar'

            formatted_user = rule['access_to'].replace('\\', '\\\\')

            self.hnas.cifs_allow_access(hnas_id, formatted_user,
                                        permission, is_snapshot=is_snapshot)

            LOG.debug("Added %(rule)s rule for user/group %(user)s "
                      "to %(entity_type)s %(share)s.",
                      {'rule': rule['access_level'],
                       'user': rule['access_to'],
                       'entity_type': entity_type,
                       'share': share_or_snapshot['id']})

    def _cifs_deny_access(self, share_or_snapshot, hnas_id, delete_rules,
                          is_snapshot=False):
        if is_snapshot:
            entity_type = "snapshot"
            share_proto = share_or_snapshot['share']['share_proto']
        else:
            entity_type = "share"
            share_proto = share_or_snapshot['share_proto']

        for rule in delete_rules:
            if rule['access_type'].lower() != 'user':
                LOG.warning('Only USER access type is allowed for '
                            'CIFS. %(entity_type)s '
                            'provided %(share)s with '
                            'protocol %(proto)s.',
                            {'entity_type': entity_type.capitalize(),
                             'share': share_or_snapshot['id'],
                             'proto': share_proto})
                continue

            formatted_user = rule['access_to'].replace('\\', '\\\\')

            self.hnas.cifs_deny_access(hnas_id, formatted_user,
                                       is_snapshot=is_snapshot)

            LOG.debug("Access denied for user/group %(user)s "
                      "to %(entity_type)s %(share)s.",
                      {'user': rule['access_to'],
                       'entity_type': entity_type,
                       'share': share_or_snapshot['id']})

    def _clean_cifs_access_list(self, hnas_id, is_snapshot=False):
        permission_list = self.hnas.list_cifs_permissions(hnas_id)

        for permission in permission_list:
            formatted_user = r'"\{1}{0}\{1}"'.format(permission[0], '"')
            self.hnas.cifs_deny_access(hnas_id, formatted_user,
                                       is_snapshot=is_snapshot)

    def create_share(self, context, share, share_server=None):
        r"""Creates share.

        :param context: The `context.RequestContext` object for the request
        :param share: Share that will be created.
        :param share_server: Data structure with share server information.
            Not used by this driver.
        :returns: Returns a list of dicts containing the EVS IP concatenated
            with the path of share in the filesystem.

            Example for NFS::

            [

              {

                'path': '172.24.44.10:/shares/id',
                'metadata': {},
                'is_admin_only': False

              },

              {

                'path': '192.168.0.10:/shares/id',
                'metadata': {},
                'is_admin_only': True

              }

            ]

            Example for CIFS::

            [

              {

                'path': '\\172.24.44.10\id',
                'metadata': {},
                'is_admin_only': False

              },

              {

                'path': '\\192.168.0.10\id',
                'metadata': {},
                'is_admin_only': True

              }

            ]

        """
        LOG.debug("Creating share in HNAS: %(shr)s.", {'shr': share['id']})

        self._check_protocol(share['id'], share['share_proto'])

        export_list = self._create_share(share['id'], share['size'],
                                         share['share_proto'])

        LOG.debug("Share %(share)s created successfully on path(s): "
                  "%(paths)s.",
                  {'paths': ', '.join([x['path'] for x in export_list]),
                   'share': share['id']})

        return export_list

    def delete_share(self, context, share, share_server=None):
        """Deletes share.

        :param context: The `context.RequestContext` object for the request
        :param share: Share that will be deleted.
        :param share_server: Data structure with share server information.
            Not used by this driver.
        """
        hnas_share_id = self._get_hnas_share_id(share['id'])

        LOG.debug("Deleting share in HNAS: %(shr)s.",
                  {'shr': share['id']})

        self._delete_share(hnas_share_id, share['share_proto'])

        LOG.debug("Export and share successfully deleted: %(shr)s.",
                  {'shr': share['id']})

    def create_snapshot(self, context, snapshot, share_server=None):
        """Creates snapshot.

        :param context: The `context.RequestContext` object for the request
        :param snapshot: Snapshot that will be created.
        :param share_server: Data structure with share server information.
            Not used by this driver.
        """
        hnas_share_id = self._get_hnas_share_id(snapshot['share_id'])

        LOG.debug("The snapshot of share %(snap_share_id)s will be created "
                  "with id %(snap_id)s.",
                  {'snap_share_id': snapshot['share_id'],
                   'snap_id': snapshot['id']})

        export_locations = self._create_snapshot(hnas_share_id, snapshot)
        LOG.info("Snapshot %(id)s successfully created.",
                 {'id': snapshot['id']})

        output = {
            'provider_location': os.path.join(
                '/snapshots', hnas_share_id, snapshot['id'])
        }

        if export_locations:
            output['export_locations'] = export_locations

        return output

    def delete_snapshot(self, context, snapshot, share_server=None):
        """Deletes snapshot.

        :param context: The `context.RequestContext` object for the request
        :param snapshot: Snapshot that will be deleted.
        :param share_server: Data structure with share server information.
            Not used by this driver.
        """
        hnas_share_id = self._get_hnas_share_id(snapshot['share_id'])
        hnas_snapshot_id = self._get_hnas_snapshot_id(snapshot)

        LOG.debug("The snapshot %(snap_id)s will be deleted. The related "
                  "share ID is %(snap_share_id)s.",
                  {'snap_id': snapshot['id'],
                   'snap_share_id': snapshot['share_id']})

        self._delete_snapshot(snapshot['share'],
                              hnas_share_id, hnas_snapshot_id)

        LOG.info("Snapshot %(id)s successfully deleted.",
                 {'id': snapshot['id']})

    def create_share_from_snapshot(self, context, share, snapshot,
                                   share_server=None, parent_share=None):
        r"""Creates a new share from snapshot.

        :param context: The `context.RequestContext` object for the request
        :param share: Information about the new share.
        :param snapshot: Information about the snapshot that will be copied
            to new share.
        :param share_server: Data structure with share server information.
            Not used by this driver.
        :returns: Returns a list of dicts containing the EVS IP concatenated
            with the path of share in the filesystem.

            Example for NFS::

            [

              {

                'path': '172.24.44.10:/shares/id',
                'metadata': {},
                'is_admin_only': False

              },

              {

                'path': '192.168.0.10:/shares/id',
                'metadata': {},
                'is_admin_only': True

              }

            ]

            Example for CIFS::

            [

              {

                'path': '\\172.24.44.10\id',
                'metadata': {},
                'is_admin_only': False

              },

              {

                'path': '\\192.168.0.10\id',
                'metadata': {},
                'is_admin_only': True

              }

            ]

        """
        LOG.debug("Creating a new share from snapshot: %(ss_id)s.",
                  {'ss_id': snapshot['id']})

        hnas_src_share_id = self._get_hnas_share_id(snapshot['share_id'])
        hnas_src_snap_id = self._get_hnas_snapshot_id(snapshot)

        export_list = self._create_share_from_snapshot(
            share, hnas_src_share_id, hnas_src_snap_id)

        LOG.debug("Share %(share)s created successfully on path(s): "
                  "%(paths)s.",
                  {'paths': ', '.join([x['path'] for x in export_list]),
                   'share': share['id']})
        return export_list

    def ensure_share(self, context, share, share_server=None):
        r"""Ensure that share is exported.

        :param context: The `context.RequestContext` object for the request
        :param share: Share that will be checked.
        :param share_server: Data structure with share server information.
            Not used by this driver.
        :returns: Returns a list of dicts containing the EVS IP concatenated
            with the path of share in the filesystem.

            Example for NFS::

            [

              {

                'path': '172.24.44.10:/shares/id',
                'metadata': {},
                'is_admin_only': False

              },

              {

                'path': '192.168.0.10:/shares/id',
                'metadata': {},
                'is_admin_only': True

              }

            ]

            Example for CIFS::

            [

              {

                'path': '\\172.24.44.10\id',
                'metadata': {},
                'is_admin_only': False

              },

              {

                'path': '\\192.168.0.10\id',
                'metadata': {},
                'is_admin_only': True

              }

            ]

        """
        LOG.debug("Ensuring share in HNAS: %(shr)s.", {'shr': share['id']})

        hnas_share_id = self._get_hnas_share_id(share['id'])

        export_list = self._ensure_share(share, hnas_share_id)

        LOG.debug("Share ensured in HNAS: %(shr)s, protocol %(proto)s.",
                  {'shr': share['id'], 'proto': share['share_proto']})
        return export_list

    def extend_share(self, share, new_size, share_server=None):
        """Extends a share to new size.

        :param share: Share that will be extended.
        :param new_size: New size of share.
        :param share_server: Data structure with share server information.
            Not used by this driver.
        """
        hnas_share_id = self._get_hnas_share_id(share['id'])

        LOG.debug("Expanding share in HNAS: %(shr_id)s.",
                  {'shr_id': share['id']})

        self._extend_share(hnas_share_id, share, new_size)
        LOG.info("Share %(shr_id)s successfully extended to "
                 "%(shr_size)s.",
                 {'shr_id': share['id'],
                  'shr_size': six.text_type(new_size)})

    # TODO(alyson): Implement in DHSS = true mode
    def get_network_allocations_number(self):
        """Track allocations_number in DHSS = true.

        When using the setting driver_handles_share_server = false
        does not require to track allocations_number because we do not handle
        network stuff.
        """
        return 0

    def _update_share_stats(self, data=None):
        """Updates the Capability of Backend."""
        LOG.debug("Updating Backend Capability Information - Hitachi HNAS.")

        self._check_fs_mounted()

        total_space, free_space, dedupe = self.hnas.get_stats()

        reserved = self.configuration.safe_get('reserved_share_percentage')
        reserved_snapshot = self.configuration.safe_get(
            'reserved_share_from_snapshot_percentage') or reserved

        data = {
            'share_backend_name': self.backend_name,
            'driver_handles_share_servers': self.driver_handles_share_servers,
            'vendor_name': 'Hitachi',
            'driver_version': '4.0.0',
            'storage_protocol': 'NFS_CIFS',
            'total_capacity_gb': total_space,
            'free_capacity_gb': free_space,
            'reserved_percentage': reserved,
            'reserved_snapshot_percentage': reserved_snapshot,
            'qos': False,
            'thin_provisioning': True,
            'dedupe': dedupe,
            'revert_to_snapshot_support': True,
            'mount_snapshot_support': True,
        }

        LOG.info("HNAS Capabilities: %(data)s.",
                 {'data': six.text_type(data)})

        super(HitachiHNASDriver, self)._update_share_stats(data)

    def manage_existing(self, share, driver_options):
        r"""Manages a share that exists on backend.

        :param share: Share that will be managed.
        :param driver_options: Empty dict or dict with 'volume_id' option.
        :returns: Returns a dict with size of the share managed and a list of
            dicts containing its export locations.

            Example for NFS::

            {

              'size': 10,
              'export_locations': [

                {

                  'path': '172.24.44.10:/shares/id',
                  'metadata': {},
                  'is_admin_only': False

                },

                {

                  'path': '192.168.0.10:/shares/id',
                  'metadata': {},
                  'is_admin_only': True

                }

              ]

            }

            Example for CIFS::

            {

              'size': 10,
              'export_locations': [

                {

                  'path': '\\172.24.44.10\id',
                  'metadata': {},
                  'is_admin_only': False

                },

                {

                  'path': '\\192.168.0.10\id',
                  'metadata': {},
                  'is_admin_only': True

                }

              ]

            }

        """
        hnas_share_id = self._get_hnas_share_id(share['id'])

        # Make sure returned value is the same as provided,
        # confirming it does not exist.
        if hnas_share_id != share['id']:
            msg = _("Share ID %s already exists, cannot manage.") % share['id']
            raise exception.HNASBackendException(msg=msg)

        self._check_protocol(share['id'], share['share_proto'])

        if share['share_proto'].lower() == 'nfs':
            # 10.0.0.1:/shares/example
            LOG.info("Share %(shr_path)s will be managed with ID "
                     "%(shr_id)s.",
                     {'shr_path': share['export_locations'][0]['path'],
                      'shr_id': share['id']})

            old_path_info = share['export_locations'][0]['path'].split(
                ':/shares/')

            if len(old_path_info) == 2:
                evs_ip = old_path_info[0]
                hnas_share_id = old_path_info[1]
            else:
                msg = _("Incorrect path. It should have the following format: "
                        "IP:/shares/share_id.")
                raise exception.ShareBackendException(msg=msg)
        else:  # then its CIFS
            # \\10.0.0.1\example
            old_path = share['export_locations'][0]['path'].split('\\')

            if len(old_path) == 4:
                evs_ip = old_path[2]
                hnas_share_id = old_path[3]
            else:
                msg = _("Incorrect path. It should have the following format: "
                        "\\\\IP\\share_id.")
                raise exception.ShareBackendException(msg=msg)

        if evs_ip != self.hnas_evs_ip:
            msg = _("The EVS IP %(evs)s is not "
                    "configured.") % {'evs': evs_ip}
            raise exception.ShareBackendException(msg=msg)

        if self.backend_name not in share['host']:
            msg = _("The backend passed in the host parameter (%(shr)s) is "
                    "not configured.") % {'shr': share['host']}
            raise exception.ShareBackendException(msg=msg)

        output = self._manage_existing(share, hnas_share_id)
        self.private_storage.update(
            share['id'], {'hnas_id': hnas_share_id})

        LOG.debug("HNAS ID %(hnas_id)s has been saved to private storage for "
                  "Share ID %(share_id)s", {'hnas_id': hnas_share_id,
                                            'share_id': share['id']})

        LOG.info("Share %(shr_path)s was successfully managed with ID "
                 "%(shr_id)s.",
                 {'shr_path': share['export_locations'][0]['path'],
                  'shr_id': share['id']})

        return output

    def unmanage(self, share):
        """Unmanages a share.

        :param share: Share that will be unmanaged.
        """
        self.private_storage.delete(share['id'])

        if len(share['export_locations']) == 0:
            LOG.info("The share with ID %(shr_id)s is no longer being "
                     "managed.", {'shr_id': share['id']})
        else:
            LOG.info("The share with current path %(shr_path)s and ID "
                     "%(shr_id)s is no longer being managed.",
                     {'shr_path': share['export_locations'][0]['path'],
                      'shr_id': share['id']})

    def shrink_share(self, share, new_size, share_server=None):
        """Shrinks a share to new size.

        :param share: Share that will be shrunk.
        :param new_size: New size of share.
        :param share_server: Data structure with share server information.
            Not used by this driver.
        """
        hnas_share_id = self._get_hnas_share_id(share['id'])

        LOG.debug("Shrinking share in HNAS: %(shr_id)s.",
                  {'shr_id': share['id']})

        self._shrink_share(hnas_share_id, share, new_size)
        LOG.info("Share %(shr_id)s successfully shrunk to "
                 "%(shr_size)sG.",
                 {'shr_id': share['id'],
                  'shr_size': six.text_type(new_size)})

    def revert_to_snapshot(self, context, snapshot, share_access_rules,
                           snapshot_access_rules, share_server=None):
        """Reverts a share to a given snapshot.

        :param context: The `context.RequestContext` object for the request
        :param snapshot: The snapshot to which the share is to be reverted to.
        :param share_access_rules: List of all access rules for the affected
            share. Not used by this driver.
        :param snapshot_access_rules: List of all access rules for the affected
            snapshot. Not used by this driver.
        :param share_server: Data structure with share server information.
            Not used by this driver.
        """

        hnas_share_id = self._get_hnas_share_id(snapshot['share_id'])

        hnas_snapshot_id = self._get_hnas_snapshot_id(snapshot)

        self._ensure_snapshot(snapshot, hnas_snapshot_id)

        dest_path = os.path.join('/shares', hnas_share_id)
        src_path = os.path.join('/snapshots', hnas_share_id, hnas_snapshot_id)

        self.hnas.tree_delete(dest_path)

        self.hnas.vvol_create(hnas_share_id)

        self.hnas.quota_add(hnas_share_id, snapshot['size'])

        try:
            self.hnas.tree_clone(src_path, dest_path)
        except exception.HNASNothingToCloneException:
            LOG.warning("Source directory is empty, creating an empty "
                        "directory.")

        LOG.info("Share %(share)s successfully reverted to snapshot "
                 "%(snapshot)s.", {'share': snapshot['share_id'],
                                   'snapshot': snapshot['id']})

    def _get_hnas_share_id(self, share_id):
        hnas_id = self.private_storage.get(share_id, 'hnas_id')

        if hnas_id is None:
            hnas_id = share_id

        LOG.debug("Share ID is %(shr_id)s and respective HNAS ID "
                  "is %(hnas_id)s.", {'shr_id': share_id,
                                      'hnas_id': hnas_id})

        return hnas_id

    def _get_hnas_snapshot_id(self, snapshot):
        hnas_snapshot_id = snapshot['id']

        if snapshot['provider_location']:
            LOG.debug("Snapshot %(snap_id)s with provider_location: "
                      "%(p_loc)s.",
                      {'snap_id': hnas_snapshot_id,
                       'p_loc': snapshot['provider_location']})
            hnas_snapshot_id = snapshot['provider_location'].split('/')[-1]

        return hnas_snapshot_id

    def _create_share(self, share_id, share_size, share_proto):
        """Creates share.

        Creates a virtual-volume, adds a quota limit and exports it.
        :param share_id: manila's database ID of share that will be created.
        :param share_size: Size limit of share.
        :param share_proto: Protocol of share that will be created
            (NFS or CIFS)
        :returns: Returns a list of dicts containing the new share's export
            locations.

        """
        self._check_fs_mounted()

        self.hnas.vvol_create(share_id)

        self.hnas.quota_add(share_id, share_size)

        LOG.debug("Share created with id %(shr)s, size %(size)sG.",
                  {'shr': share_id, 'size': share_size})

        self._create_export(share_id, share_proto)

        export_list = self._get_export_locations(share_proto, share_id)
        return export_list

    def _create_export(self, share_id, share_proto, snapshot_id=None):
        try:
            if share_proto.lower() == 'nfs':
                # Create NFS export
                self.hnas.nfs_export_add(share_id, snapshot_id=snapshot_id)
                LOG.debug("NFS Export created to %(shr)s.",
                          {'shr': share_id})
            else:
                # Create CIFS share with vvol path
                self.hnas.cifs_share_add(share_id, snapshot_id=snapshot_id)
                LOG.debug("CIFS share created to %(shr)s.",
                          {'shr': share_id})
        except exception.HNASBackendException:
            with excutils.save_and_reraise_exception():
                if snapshot_id is None:
                    self.hnas.vvol_delete(share_id)

    def _check_fs_mounted(self):
        mounted = self.hnas.check_fs_mounted()
        if not mounted:
            msg = _("Filesystem %s is not mounted.") % self.fs_name
            raise exception.HNASBackendException(msg=msg)

    def _ensure_share(self, share, hnas_share_id):
        """Ensure that share is exported.

        :param share: Share that will be checked.
        :param hnas_share_id: HNAS ID of share that will be checked.
        :returns: Returns a list of dicts containing the share's export
            locations.

        """
        self._check_protocol(share['id'], share['share_proto'])
        self._check_fs_mounted()

        self.hnas.check_vvol(hnas_share_id)
        self.hnas.check_quota(hnas_share_id)

        if share['share_proto'].lower() == 'nfs':
            self.hnas.check_export(hnas_share_id)
        else:
            self.hnas.check_cifs(hnas_share_id)

        export_list = self._get_export_locations(
            share['share_proto'], hnas_share_id)

        return export_list

    def _shrink_share(self, hnas_share_id, share, new_size):
        """Shrinks a share to new size.

        :param hnas_share_id: HNAS ID of share that will be shrunk.
        :param share: model of share that will be shrunk.
        :param new_size: New size of share after shrink operation.
        """
        self._ensure_share(share, hnas_share_id)

        usage = self.hnas.get_share_usage(hnas_share_id)

        LOG.debug("Usage space in share %(share)s: %(usage)sG",
                  {'share': share['id'], 'usage': usage})

        if new_size > usage:
            self.hnas.modify_quota(hnas_share_id, new_size)
        else:
            raise exception.ShareShrinkingPossibleDataLoss(
                share_id=share['id'])

    def _extend_share(self, hnas_share_id, share, new_size):
        """Extends a share to new size.

        :param hnas_share_id: HNAS ID of share that will be extended.
        :param share: model of share that will be extended.
        :param new_size: New size of share after extend operation.
        """
        self._ensure_share(share, hnas_share_id)

        old_size = share['size']
        available_space = self.hnas.get_stats()[1]

        LOG.debug("Available space in filesystem: %(space)sG.",
                  {'space': available_space})

        if (new_size - old_size) < available_space:
            self.hnas.modify_quota(hnas_share_id, new_size)
        else:
            msg = (_("Share %s cannot be extended due to insufficient space.")
                   % share['id'])
            raise exception.HNASBackendException(msg=msg)

    def _delete_share(self, hnas_share_id, share_proto):
        """Deletes share.

        It uses tree-delete-job-submit to format and delete virtual-volumes.
        Quota is deleted with virtual-volume.
        :param hnas_share_id: HNAS ID of share that will be deleted.
        :param share_proto: Protocol of share that will be deleted.
        """
        self._check_fs_mounted()

        if share_proto.lower() == 'nfs':
            self.hnas.nfs_export_del(hnas_share_id)
        elif share_proto.lower() == 'cifs':
            self.hnas.cifs_share_del(hnas_share_id)
        self.hnas.vvol_delete(hnas_share_id)

    def _manage_existing(self, share, hnas_share_id):
        """Manages a share that exists on backend.

        :param share: share that will be managed.
        :param hnas_share_id: HNAS ID of share that will be managed.
        :returns: Returns a dict with size of the share managed and a list of
            dicts containing its export locations.
        """
        self._ensure_share(share, hnas_share_id)

        share_size = self.hnas.get_share_quota(hnas_share_id)
        if share_size is None:
            msg = (_("The share %s trying to be managed does not have a "
                     "quota limit, please set it before manage.")
                   % share['id'])
            raise exception.ManageInvalidShare(reason=msg)

        export_list = self._get_export_locations(
            share['share_proto'], hnas_share_id)

        return {'size': share_size, 'export_locations': export_list}

    def _create_snapshot(self, hnas_share_id, snapshot):
        """Creates a snapshot of share.

        It copies the directory and all files to a new directory inside
        /snapshots/share_id/.
        :param hnas_share_id: HNAS ID of share for snapshot.
        :param snapshot: Snapshot that will be created.
        """
        self._ensure_share(snapshot['share'], hnas_share_id)
        saved_list = []

        share_proto = snapshot['share']['share_proto']
        self._check_protocol(snapshot['share_id'], share_proto)

        if share_proto.lower() == 'nfs':
            saved_list = self.hnas.get_nfs_host_list(hnas_share_id)
            new_list = []
            for access in saved_list:
                for rw in ('read_write', 'readwrite', 'rw'):
                    access = access.replace(rw, 'ro')
                new_list.append(access)
            self.hnas.update_nfs_access_rule(new_list, share_id=hnas_share_id)
        else:  # CIFS
            if (self.hnas.is_cifs_in_use(hnas_share_id) and
                    not self.cifs_snapshot):
                msg = _("CIFS snapshot when share is mounted is disabled. "
                        "Set hitachi_hnas_allow_cifs_snapshot_while_mounted to"
                        " True or unmount the share to take a snapshot.")
                raise exception.ShareBackendException(msg=msg)

        src_path = os.path.join('/shares', hnas_share_id)
        dest_path = os.path.join('/snapshots', hnas_share_id, snapshot['id'])
        try:
            self.hnas.tree_clone(src_path, dest_path)
        except exception.HNASNothingToCloneException:
            LOG.warning("Source directory is empty, creating an empty "
                        "directory.")
            self.hnas.create_directory(dest_path)
        finally:
            if share_proto.lower() == 'nfs':
                self.hnas.update_nfs_access_rule(saved_list,
                                                 share_id=hnas_share_id)

        export_locations = []

        if snapshot['share'].get('mount_snapshot_support'):
            self._create_export(hnas_share_id, share_proto,
                                snapshot_id=snapshot['id'])
            export_locations = self._get_export_locations(
                share_proto, snapshot['id'], is_snapshot=True)

        return export_locations

    def _delete_snapshot(self, share, hnas_share_id, snapshot_id):
        """Deletes snapshot.

        It receives the hnas_share_id only to join the path for snapshot.
        :param hnas_share_id: HNAS ID of share from which snapshot was taken.
        :param snapshot_id: ID of snapshot.
        """
        self._check_fs_mounted()
        share_proto = share['share_proto']

        if share.get('mount_snapshot_support'):
            if share_proto.lower() == 'nfs':
                self.hnas.nfs_export_del(snapshot_id=snapshot_id)
            elif share_proto.lower() == 'cifs':
                self.hnas.cifs_share_del(snapshot_id)

        path = os.path.join('/snapshots', hnas_share_id, snapshot_id)
        self.hnas.tree_delete(path)
        path = os.path.join('/snapshots', hnas_share_id)
        self.hnas.delete_directory(path)

    def _create_share_from_snapshot(self, share, src_hnas_share_id,
                                    hnas_snapshot_id):
        """Creates a new share from snapshot.

        It copies everything from snapshot directory to a new vvol,
        set a quota limit for it and export.
        :param share: a dict from new share.
        :param src_hnas_share_id: HNAS ID of share from which snapshot was
            taken.
        :param hnas_snapshot_id: HNAS ID from snapshot that will be copied to
            new share.
        :returns: Returns a list of dicts containing the new share's export
            locations.
        """
        dest_path = os.path.join('/shares', share['id'])
        src_path = os.path.join('/snapshots', src_hnas_share_id,
                                hnas_snapshot_id)

        # Before copying everything to new vvol, we need to create it,
        # because we only can transform an empty directory into a vvol.

        self._check_fs_mounted()

        self.hnas.vvol_create(share['id'])

        self.hnas.quota_add(share['id'], share['size'])

        try:
            self.hnas.tree_clone(src_path, dest_path)
        except exception.HNASNothingToCloneException:
            LOG.warning("Source directory is empty, exporting "
                        "directory.")

        self._check_protocol(share['id'], share['share_proto'])

        try:
            if share['share_proto'].lower() == 'nfs':
                self.hnas.nfs_export_add(share['id'])

            else:
                self.hnas.cifs_share_add(share['id'])
        except exception.HNASBackendException:
            with excutils.save_and_reraise_exception():
                self.hnas.vvol_delete(share['id'])

        return self._get_export_locations(
            share['share_proto'], share['id'])

    def _check_protocol(self, share_id, protocol):
        if protocol.lower() not in ('nfs', 'cifs'):
            msg = _("Only NFS or CIFS protocol are currently supported. "
                    "Share provided %(share)s with protocol "
                    "%(proto)s.") % {'share': share_id,
                                     'proto': protocol}
            raise exception.ShareBackendException(msg=msg)

    def _get_export_locations(self, share_proto, hnas_id, is_snapshot=False):
        export_list = []
        for ip in (self.hnas_evs_ip, self.hnas_admin_network_ip):
            if ip:
                path = self._get_export_path(ip, share_proto, hnas_id,
                                             is_snapshot)
                export_list.append({
                    "path": path,
                    "is_admin_only": ip == self.hnas_admin_network_ip,
                    "metadata": {},
                })
        return export_list

    def _get_export_path(self, ip, share_proto, hnas_id, is_snapshot):
        r"""Gets and returns export path.

        :param ip: IP from HNAS EVS configured.
        :param share_proto: Share or snapshot protocol (NFS or CIFS).
        :param hnas_id: Entity ID in HNAS, it can be the ID from a share or
            a snapshot.
        :param is_snapshot: Boolean to determine if export is related to a
            share or a snapshot.
        :return: Complete export path, for example:
            - In NFS:
                SHARE: 172.24.44.10:/shares/id
                SNAPSHOT: 172.24.44.10:/snapshots/id
            - In CIFS:
                SHARE and SNAPSHOT: \\172.24.44.10\id
        """
        if share_proto.lower() == 'nfs':
            if is_snapshot:
                path = os.path.join('/snapshots', hnas_id)
            else:
                path = os.path.join('/shares', hnas_id)
            export = ':'.join((ip, path))
        else:
            export = r'\\%s\%s' % (ip, hnas_id)
        return export

    def _ensure_snapshot(self, snapshot, hnas_snapshot_id):
        """Ensure that snapshot is exported.

        :param snapshot: Snapshot that will be checked.
        :param hnas_snapshot_id: HNAS ID of snapshot that will be checked.

        :returns: Returns a list of dicts containing the snapshot's export
            locations or None if mount_snapshot_support is False.
        """
        self._check_protocol(snapshot['share_id'],
                             snapshot['share']['share_proto'])
        self._check_fs_mounted()

        self.hnas.check_directory(snapshot['provider_location'])

        export_list = None
        if snapshot['share'].get('mount_snapshot_support'):
            if snapshot['share']['share_proto'].lower() == 'nfs':
                self.hnas.check_export(hnas_snapshot_id, is_snapshot=True)
            else:
                self.hnas.check_cifs(hnas_snapshot_id)

            export_list = self._get_export_locations(
                snapshot['share']['share_proto'],
                hnas_snapshot_id,
                is_snapshot=True)

        return export_list

    def ensure_snapshot(self, context, snapshot, share_server=None):
        r"""Ensure that snapshot is exported.

        :param context: The `context.RequestContext` object for the request.
        :param snapshot: Snapshot that will be checked.
        :param share_server: Data structure with share server information.
            Not used by this driver.

        :returns: Returns a list of dicts containing the EVS IP concatenated
            with the path of snapshot in the filesystem or None if
            mount_snapshot_support is False.

            Example for NFS::

            [

              {

                'path': '172.24.44.10:/snapshots/id',
                'metadata': {},
                'is_admin_only': False

              },

              {

                'path': '192.168.0.10:/snapshots/id',
                'metadata': {},
                'is_admin_only': True

              }

            ]

            Example for CIFS::

            [

              {

                'path': '\\172.24.44.10\id',
                'metadata': {},
                'is_admin_only': False

              },

              {

                'path': '\\192.168.0.10\id',
                'metadata': {},
                'is_admin_only': True

              }

            ]

        """
        LOG.debug("Ensuring snapshot in HNAS: %(snap)s.",
                  {'snap': snapshot['id']})

        hnas_snapshot_id = self._get_hnas_snapshot_id(snapshot)

        export_list = self._ensure_snapshot(snapshot, hnas_snapshot_id)

        LOG.debug("Snapshot ensured in HNAS: %(snap)s, protocol %(proto)s.",
                  {'snap': snapshot['id'],
                   'proto': snapshot['share']['share_proto']})
        return export_list

    def manage_existing_snapshot(self, snapshot, driver_options):
        """Manages a snapshot that exists only in HNAS.

        The snapshot to be managed should be in the path
        /snapshots/SHARE_ID/SNAPSHOT_ID. Also, the size of snapshot should be
        provided as --driver_options size=<size>.
        :param snapshot: snapshot that will be managed.
        :param driver_options: expects only one key 'size'. It must be
        provided in order to manage a snapshot.

        :returns: Returns a dict with size of snapshot managed
        """
        try:
            snapshot_size = int(driver_options.get("size", 0))
        except (ValueError, TypeError):
            msg = _("The size in driver options to manage snapshot "
                    "%(snap_id)s should be an integer, in format "
                    "driver-options size=<SIZE>. Value passed: "
                    "%(size)s.") % {'snap_id': snapshot['id'],
                                    'size': driver_options.get("size")}
            raise exception.ManageInvalidShareSnapshot(reason=msg)

        if snapshot_size == 0:
            msg = _("Snapshot %(snap_id)s has no size specified for manage. "
                    "Please, provide the size with parameter driver-options "
                    "size=<SIZE>.") % {'snap_id': snapshot['id']}
            raise exception.ManageInvalidShareSnapshot(reason=msg)

        hnas_share_id = self._get_hnas_share_id(snapshot['share_id'])

        LOG.debug("Path provided to manage snapshot: %(path)s.",
                  {'path': snapshot['provider_location']})

        path_info = snapshot['provider_location'].split('/')

        if len(path_info) == 4 and path_info[1] == 'snapshots':
            path_share_id = path_info[2]
            hnas_snapshot_id = path_info[3]
        else:
            msg = (_("Incorrect path %(path)s for manage snapshot "
                     "%(snap_id)s. It should have the following format: "
                     "/snapshots/SHARE_ID/SNAPSHOT_ID.") %
                   {'path': snapshot['provider_location'],
                    'snap_id': snapshot['id']})
            raise exception.ManageInvalidShareSnapshot(reason=msg)

        if hnas_share_id != path_share_id:
            msg = _("The snapshot %(snap_id)s does not belong to share "
                    "%(share_id)s.") % {'snap_id': snapshot['id'],
                                        'share_id': snapshot['share_id']}
            raise exception.ManageInvalidShareSnapshot(reason=msg)

        if not self.hnas.check_directory(snapshot['provider_location']):
            msg = _("Snapshot %(snap_id)s does not exist in "
                    "HNAS.") % {'snap_id': hnas_snapshot_id}
            raise exception.ManageInvalidShareSnapshot(reason=msg)

        try:
            self._ensure_snapshot(snapshot, hnas_snapshot_id)
        except exception.HNASItemNotFoundException:
            LOG.warning("Export does not exist for snapshot %s, "
                        "creating a new one.", snapshot['id'])
            self._create_export(hnas_share_id,
                                snapshot['share']['share_proto'],
                                snapshot_id=hnas_snapshot_id)

        output = {'size': snapshot_size}
        if snapshot['share'].get('mount_snapshot_support'):
            export_locations = self._get_export_locations(
                snapshot['share']['share_proto'],
                hnas_snapshot_id,
                is_snapshot=True)
            output['export_locations'] = export_locations

        LOG.info("Snapshot %(snap_path)s for share %(shr_id)s was "
                 "successfully managed with ID %(snap_id)s.",
                 {'snap_path': snapshot['provider_location'],
                  'shr_id': snapshot['share_id'],
                  'snap_id': snapshot['id']})

        return output

    def unmanage_snapshot(self, snapshot):
        """Unmanage a share snapshot

        :param snapshot: Snapshot that will be unmanaged.
        """
        LOG.info("The snapshot with ID %(snap_id)s from share "
                 "%(share_id)s is no longer being managed by Manila. "
                 "However, it is not deleted and can be found in HNAS.",
                 {'snap_id': snapshot['id'],
                  'share_id': snapshot['share_id']})

    def snapshot_update_access(self, context, snapshot, access_rules,
                               add_rules, delete_rules, share_server=None):
        """Update access rules for given snapshot.

        Drivers should support 2 different cases in this method:
        1. Recovery after error - 'access_rules' contains all access rules,
        'add_rules' and 'delete_rules' shall be empty. Driver should clear any
        existent access rules and apply all access rules for given snapshot.
        This recovery is made at driver start up.

        2. Adding/Deleting of several access rules - 'access_rules' contains
        all access rules, 'add_rules' and 'delete_rules' contain rules which
        should be added/deleted. Driver can ignore rules in 'access_rules' and
        apply only rules from 'add_rules' and 'delete_rules'. All snapshots
        rules should be read only.

        :param context: Current context
        :param snapshot: Snapshot model with snapshot data.
        :param access_rules: All access rules for given snapshot
        :param add_rules: Empty List or List of access rules which should be
               added. access_rules already contains these rules.
        :param delete_rules: Empty List or List of access rules which should be
               removed. access_rules doesn't contain these rules.
        :param share_server: None or Share server model
        """
        hnas_snapshot_id = self._get_hnas_snapshot_id(snapshot)

        self._ensure_snapshot(snapshot, hnas_snapshot_id)

        access_rules, add_rules, delete_rules = utils.change_rules_to_readonly(
            access_rules, add_rules, delete_rules)

        if snapshot['share']['share_proto'].lower() == 'nfs':
            host_list = []

            for rule in access_rules:
                if rule['access_type'].lower() != 'ip':
                    msg = _("Only IP access type currently supported for NFS. "
                            "Snapshot provided %(snapshot)s with rule type "
                            "%(type)s.") % {'snapshot': snapshot['id'],
                                            'type': rule['access_type']}
                    raise exception.InvalidSnapshotAccess(reason=msg)

                host_list.append(rule['access_to'] + '(ro)')

            self.hnas.update_nfs_access_rule(host_list,
                                             snapshot_id=hnas_snapshot_id)

            if host_list:
                LOG.debug("Snapshot %(snapshot)s has the rules: %(rules)s",
                          {'snapshot': snapshot['id'],
                           'rules': ', '.join(host_list)})
            else:
                LOG.debug("Snapshot %(snapshot)s has no rules.",
                          {'snapshot': snapshot['id']})
        else:
            if not (add_rules or delete_rules):
                # cifs recovery mode
                self._clean_cifs_access_list(hnas_snapshot_id,
                                             is_snapshot=True)
                self._cifs_allow_access(snapshot, hnas_snapshot_id,
                                        access_rules, is_snapshot=True)
            else:
                self._cifs_deny_access(snapshot, hnas_snapshot_id,
                                       delete_rules, is_snapshot=True)
                self._cifs_allow_access(snapshot, hnas_snapshot_id,
                                        add_rules, is_snapshot=True)



#    Copyright 2010 OpenStack LLC
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import glob
import os


from manila import exception
from manila import test


class ExceptionTestCase(test.TestCase):
    @staticmethod
    def _raise_exc(exc):
        raise exc()

    def test_exceptions_raise(self):
        # NOTE(dprince): disable format errors since we are not passing kwargs
        self.flags(fatal_exception_format_errors=False)
        for name in dir(exception):
            exc = getattr(exception, name)
            if isinstance(exc, type):
                self.assertRaises(exc, self._raise_exc, exc)


class ProjectTestCase(test.TestCase):
    def test_all_migrations_have_downgrade(self):
        topdir = os.path.normpath(os.path.dirname(__file__) + '/../../')
        py_glob = os.path.join(topdir, "manila", "db", "sqlalchemy",
                               "migrate_repo", "versions", "*.py")
        missing_downgrade = []
        for path in glob.iglob(py_glob):
            has_upgrade = False
            has_downgrade = False
            with open(path, "r") as f:
                for line in f:
                    if 'def upgrade(' in line:
                        has_upgrade = True
                    if 'def downgrade(' in line:
                        has_downgrade = True

                if has_upgrade and not has_downgrade:
                    fname = os.path.basename(path)
                    missing_downgrade.append(fname)

        helpful_msg = ("The following migrations are missing a downgrade:"
                       "\n\t%s") % '\n\t'.join(sorted(missing_downgrade))
        self.assertTrue(not missing_downgrade, helpful_msg)



# Copyright 2019 Inspur Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
CLI helpers for Inspur InStorage
"""

import paramiko
import re
import six

from eventlet import greenthread

from oslo_concurrency import processutils
from oslo_log import log
from oslo_utils import excutils

from manila import exception
from manila.i18n import _
from manila import utils as manila_utils

LOG = log.getLogger(__name__)


class SSHRunner(object):
    """SSH runner is used to run ssh command on inspur instorage system."""

    def __init__(self, host, port, login, password, privatekey=None):
        self.host = host
        self.port = port
        self.login = login
        self.password = password
        self.privatekey = privatekey

        self.ssh_conn_timeout = 60
        self.ssh_min_pool_size = 1
        self.ssh_max_pool_size = 10

        self.sshpool = None

    def __call__(self, cmd_list, check_exit_code=True, attempts=1):
        """SSH tool"""
        manila_utils.check_ssh_injection(cmd_list)
        command = ' '.join(cmd_list)
        if not self.sshpool:
            try:
                self.sshpool = manila_utils.SSHPool(
                    self.host,
                    self.port,
                    self.ssh_conn_timeout,
                    self.login,
                    password=self.password,
                    privatekey=self.privatekey,
                    min_size=self.ssh_min_pool_size,
                    max_size=self.ssh_max_pool_size
                )
            except paramiko.SSHException:
                LOG.error("Unable to create SSHPool")
                raise
        try:
            return self._ssh_execute(self.sshpool, command,
                                     check_exit_code, attempts)
        except Exception:
            LOG.error("Error running SSH command: %s", command)
            raise

    def _ssh_execute(self, sshpool, command,
                     check_exit_code=True, attempts=1):
        try:
            with sshpool.item() as ssh:
                last_exception = None
                while attempts > 0:
                    attempts -= 1
                    try:
                        return processutils.ssh_execute(
                            ssh,
                            command,
                            check_exit_code=check_exit_code)
                    except Exception as e:
                        LOG.exception('Error has occurred')
                        last_exception = e
                        greenthread.sleep(1)

                try:
                    raise processutils.ProcessExecutionError(
                        exit_code=last_exception.exit_code,
                        stdout=last_exception.stdout,
                        stderr=last_exception.stderr,
                        cmd=last_exception.cmd)
                except AttributeError:
                    raise processutils.ProcessExecutionError(
                        exit_code=-1,
                        stdout="",
                        stderr="Error running SSH command",
                        cmd=command)

        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error("Error running SSH command: %s", command)


class CLIParser(object):
    """Parse MCS CLI output and generate iterable."""

    def __init__(self, raw, ssh_cmd=None, delim='!', with_header=True):
        super(CLIParser, self).__init__()
        if ssh_cmd:
            self.ssh_cmd = ' '.join(ssh_cmd)
        else:
            self.ssh_cmd = 'None'
        self.raw = raw
        self.delim = delim
        self.with_header = with_header
        self.result = self._parse()

    def __getitem__(self, key):
        try:
            return self.result[key]
        except KeyError:
            msg = (_('Did not find the expected key %(key)s in %(fun)s: '
                     '%(raw)s.') % {'key': key, 'fun': self.ssh_cmd,
                                    'raw': self.raw})
            raise exception.ShareBackendException(msg=msg)

    def __iter__(self):
        for a in self.result:
            yield a

    def __len__(self):
        return len(self.result)

    def _parse(self):
        def get_reader(content, delim):
            for line in content.lstrip().splitlines():
                line = line.strip()
                if line:
                    yield line.split(delim)
                else:
                    yield []

        if isinstance(self.raw, six.string_types):
            stdout, stderr = self.raw, ''
        else:
            stdout, stderr = self.raw
        reader = get_reader(stdout, self.delim)
        result = []

        if self.with_header:
            hds = tuple()
            for row in reader:
                hds = row
                break
            for row in reader:
                cur = dict()
                if len(hds) != len(row):
                    msg = (_('Unexpected CLI response: header/row mismatch. '
                             'header: %(header)s, row: %(row)s.')
                           % {'header': hds,
                              'row': row})
                    raise exception.ShareBackendException(msg=msg)
                for k, v in zip(hds, row):
                    CLIParser.append_dict(cur, k, v)
                result.append(cur)
        else:
            cur = dict()
            for row in reader:
                if row:
                    CLIParser.append_dict(cur, row[0], ' '.join(row[1:]))
                elif cur:  # start new section
                    result.append(cur)
                    cur = dict()
            if cur:
                result.append(cur)
        return result

    @staticmethod
    def append_dict(dict_, key, value):
        key, value = key.strip(), value.strip()
        obj = dict_.get(key, None)
        if obj is None:
            dict_[key] = value
        elif isinstance(obj, list):
            obj.append(value)
            dict_[key] = obj
        else:
            dict_[key] = [obj, value]
        return dict_


class InStorageSSH(object):
    """SSH interface to Inspur InStorage systems."""

    def __init__(self, ssh_runner):
        self._ssh = ssh_runner

    def _run_ssh(self, ssh_cmd):
        try:
            return self._ssh(ssh_cmd)
        except processutils.ProcessExecutionError as e:
            msg = (_('CLI Exception output:\n command: %(cmd)s\n '
                     'stdout: %(out)s\n stderr: %(err)s.') %
                   {'cmd': ssh_cmd,
                    'out': e.stdout,
                    'err': e.stderr})
            LOG.error(msg)
            raise exception.ShareBackendException(msg=msg)

    def run_ssh_inq(self, ssh_cmd, delim='!', with_header=False):
        """Run an SSH command and return parsed output."""
        raw = self._run_ssh(ssh_cmd)
        LOG.debug('Response for cmd %s is %s', ssh_cmd, raw)
        return CLIParser(raw, ssh_cmd=ssh_cmd, delim=delim,
                         with_header=with_header)

    def run_ssh_assert_no_output(self, ssh_cmd):
        """Run an SSH command and assert no output returned."""
        out, err = self._run_ssh(ssh_cmd)
        if len(out.strip()) != 0:
            msg = (_('Expected no output from CLI command %(cmd)s, '
                     'got %(out)s.') % {'cmd': ' '.join(ssh_cmd), 'out': out})
            LOG.error(msg)
            raise exception.ShareBackendException(msg=msg)

    def run_ssh_check_created(self, ssh_cmd):
        """Run an SSH command and return the ID of the created object."""
        out, err = self._run_ssh(ssh_cmd)
        try:
            match_obj = re.search(r'\[([0-9]+)\],? successfully created', out)
            return match_obj.group(1)
        except (AttributeError, IndexError):
            msg = (_('Failed to parse CLI output:\n command: %(cmd)s\n '
                     'stdout: %(out)s\n stderr: %(err)s.') %
                   {'cmd': ssh_cmd,
                    'out': out,
                    'err': err})
            LOG.error(msg)
            raise exception.ShareBackendException(msg=msg)

    def lsnode(self, node_id=None):
        with_header = True
        ssh_cmd = ['mcsinq', 'lsnode', '-delim', '!']
        if node_id:
            with_header = False
            ssh_cmd.append(node_id)
        return self.run_ssh_inq(ssh_cmd, with_header=with_header)

    def lsnaspool(self, pool_id=None):
        ssh_cmd = ['mcsinq', 'lsnaspool', '-delim', '!']
        if pool_id:
            ssh_cmd.append(pool_id)
        return self.run_ssh_inq(ssh_cmd, with_header=True)

    def lsfs(self, node_name=None, fsname=None):
        if fsname and not node_name:
            msg = _('Node name should be set when file system name is set.')
            LOG.error(msg)
            raise exception.InvalidParameterValue(msg)

        ssh_cmd = ['mcsinq', 'lsfs', '-delim', '!']
        to_append = []

        if node_name:
            to_append += ['-node', '"%s"' % node_name]

        if fsname:
            to_append += ['-name', '"%s"' % fsname]

        if not to_append:
            to_append += ['-all']

        ssh_cmd += to_append
        return self.run_ssh_inq(ssh_cmd, with_header=True)

    def addfs(self, fsname, pool_name, size, node_name):
        """Create a file system on the storage.

        :param fsname: file system name
        :param pool_name: pool in which to create the file system
        :param size: file system size in GB
        :param node_name: the primary node name
        :return:
        """

        ssh_cmd = ['mcsop', 'addfs', '-name', '"%s"' % fsname, '-pool',
                   '"%s"' % pool_name, '-size', '%dg' % size,
                   '-node', '"%s"' % node_name]
        self.run_ssh_assert_no_output(ssh_cmd)

    def rmfs(self, fsname):
        """Remove the specific file system.

        :param fsname: file system name to be removed
        :return:
        """

        ssh_cmd = ['mcsop', 'rmfs', '-name', '"%s"' % fsname]
        self.run_ssh_assert_no_output(ssh_cmd)

    def expandfs(self, fsname, size):
        """Expand the space of the specific file system.

        :param fsname: file system name
        :param size: the size(GB) to be expanded, origin + size = result
        :return:
        """

        ssh_cmd = ['mcsop', 'expandfs', '-name', '"%s"' % fsname,
                   '-size', '%dg' % size]
        self.run_ssh_assert_no_output(ssh_cmd)

    # NAS directory operation
    def lsnasdir(self, dirpath):
        """List the child directory under dirpath.

        :param dirpath: the parent directory to list with
        :return:
        """

        ssh_cmd = ['mcsinq', 'lsnasdir', '-delim', '!', '"%s"' % dirpath]
        return self.run_ssh_inq(ssh_cmd, with_header=True)

    def addnasdir(self, dirpath):
        """Create a new NAS directory indicated by dirpath."""

        ssh_cmd = ['mcsop', 'addnasdir', '"%s"' % dirpath]
        self.run_ssh_assert_no_output(ssh_cmd)

    def chnasdir(self, old_path, new_path):
        """Rename the NAS directory name."""

        ssh_cmd = ['mcsop', 'chnasdir', '-oldpath', '"%s"' % old_path,
                   '-newpath', '"%s"' % new_path]
        self.run_ssh_assert_no_output(ssh_cmd)

    def rmnasdir(self, dirpath):
        """Remove the specific dirpath."""

        ssh_cmd = ['mcsop', 'rmnasdir', '"%s"' % dirpath]
        self.run_ssh_assert_no_output(ssh_cmd)

    # NFS operation
    def rmnfs(self, share_path):
        """Remove the NFS indicated by path."""

        ssh_cmd = ['mcsop', 'rmnfs', '"%s"' % share_path]
        self.run_ssh_assert_no_output(ssh_cmd)

    def lsnfslist(self, prefix=None):
        """List NFS shares on a system."""

        ssh_cmd = ['mcsinq', 'lsnfslist', '-delim', '!']
        if prefix:
            ssh_cmd.append('"%s"' % prefix)

        return self.run_ssh_inq(ssh_cmd, with_header=True)

    def lsnfsinfo(self, share_path):
        """List a specific NFS share's information."""

        ssh_cmd = ['mcsinq', 'lsnfsinfo', '-delim', '!', '"%s"' % share_path]
        return self.run_ssh_inq(ssh_cmd, with_header=True)

    def addnfsclient(self, share_path, client_spec):
        """Add a client access rule to NFS share.

        :param share_path: the NFS share path.
        :param client_spec: IP/MASK:RIGHTS:ALL_SQUASH:ROOT_SQUASH.
        :return:
        """

        ssh_cmd = ['mcsop', 'addnfsclient', '-path', '"%s"' % share_path,
                   '-client', client_spec]
        self.run_ssh_assert_no_output(ssh_cmd)

    def chnfsclient(self, share_path, client_spec):
        """Change a NFS share's client info."""

        ssh_cmd = ['mcsop', 'chnfsclient', '-path', '"%s"' % share_path,
                   '-client', client_spec]
        self.run_ssh_assert_no_output(ssh_cmd)

    def rmnfsclient(self, share_path, client_spec):
        """Remove a client info from the NFS share."""

        # client_spec parameter for rmnfsclient is IP/MASK,
        # so we need remove the right part
        client_spec = client_spec.split(':')[0]

        ssh_cmd = ['mcsop', 'rmnfsclient', '-path', '"%s"' % share_path,
                   '-client', client_spec]
        self.run_ssh_assert_no_output(ssh_cmd)

    # CIFS operation
    def lscifslist(self, filter=None):
        """List CIFS shares on the system."""

        ssh_cmd = ['mcsinq', 'lscifslist', '-delim', '!']
        if filter:
            ssh_cmd.append('"%s"' % filter)

        return self.run_ssh_inq(ssh_cmd, with_header=True)

    def lscifsinfo(self, share_name):
        """List a specific CIFS share's information."""

        ssh_cmd = ['mcsinq', 'lscifsinfo', '-delim', '!', '"%s"' % share_name]
        return self.run_ssh_inq(ssh_cmd, with_header=True)

    def addcifs(self, share_name, dirpath, oplocks='off'):
        """Create a CIFS share with given path."""
        ssh_cmd = ['mcsop', 'addcifs', '-name', share_name, '-path', dirpath,
                   '-oplocks', oplocks]
        self.run_ssh_assert_no_output(ssh_cmd)

    def rmcifs(self, share_name):
        """Remove a CIFS share."""

        ssh_cmd = ['mcsop', 'rmcifs', share_name]
        self.run_ssh_assert_no_output(ssh_cmd)

    def chcifs(self, share_name, oplocks='off'):
        """Change a CIFS share's attribute.

        :param share_name: share's name
        :param oplocks: 'off' or 'on'
        :return:
        """
        ssh_cmd = ['mcsop', 'chcifs', '-name', share_name, '-oplocks', oplocks]
        self.run_ssh_assert_no_output(ssh_cmd)

    def addcifsuser(self, share_name, rights):
        """Add a user access rule to CIFS share.

        :param share_name: share's name
        :param rights: [LU|LG]:xxx:[rw|ro]
        :return:
        """
        ssh_cmd = ['mcsop', 'addcifsuser', '-name', share_name,
                   '-rights', rights]
        self.run_ssh_assert_no_output(ssh_cmd)

    def chcifsuser(self, share_name, rights):
        """Change a user access rule."""

        ssh_cmd = ['mcsop', 'chcifsuser', '-name', share_name,
                   '-rights', rights]
        self.run_ssh_assert_no_output(ssh_cmd)

    def rmcifsuser(self, share_name, rights):
        """Remove CIFS user from a CIFS share."""

        # the rights parameter for rmcifsuser is LU:NAME
        rights = ':'.join(rights.split(':')[0:-1])

        ssh_cmd = ['mcsop', 'rmcifsuser', '-name', share_name,
                   '-rights', rights]
        self.run_ssh_assert_no_output(ssh_cmd)

    # NAS port ip
    def lsnasportip(self):
        """List NAS service port ip address."""

        ssh_cmd = ['mcsinq', 'lsnasportip', '-delim', '!']
        return self.run_ssh_inq(ssh_cmd, with_header=True)
